import sys
import os
import requests

import torch
import numpy as np

import matplotlib.pyplot as plt
from PIL import Image

import visionsuite.cores.masked_autoencoder.mae.models_mae as models_mae
from visionsuite.utils.download import download_weights_from_url

imagenet_mean = np.array([0.485, 0.456, 0.406])
imagenet_std = np.array([0.229, 0.224, 0.225])

def prepare_model(chkpt_dir, arch='mae_vit_large_patch16'):
    # build model
    model = getattr(models_mae, arch)()
    # load model
    checkpoint = torch.load(chkpt_dir, map_location='cpu')
    msg = model.load_state_dict(checkpoint['model'], strict=False)
    print(msg)
    return model

def run_one_image(img, model, output_filename):
    x = torch.tensor(img)

    # make it a batch-like
    x = x.unsqueeze(dim=0)
    x = torch.einsum('nhwc->nchw', x)

    # run MAE
    loss, y, mask = model(x.float(), mask_ratio=0.75)
    y = model.unpatchify(y)
    y = torch.einsum('nchw->nhwc', y).detach().cpu()

    # visualize the mask
    mask = mask.detach()
    mask = mask.unsqueeze(-1).repeat(1, 1, model.patch_embed.patch_size[0]**2 *3)  # (N, H*W, p*p*3)
    mask = model.unpatchify(mask)  # 1 is removing, 0 is keeping
    mask = torch.einsum('nchw->nhwc', mask).detach().cpu()
    
    x = torch.einsum('nchw->nhwc', x)

    # masked image
    im_masked = x * (1 - mask)

    # MAE reconstruction pasted with visible patches
    im_paste = x * (1 - mask) + y * mask

    # make the plt figure larger
    plt.rcParams['figure.figsize'] = [24, 24]

    plt.subplot(1, 4, 1)
    plt.imshow(torch.clip((x[0] * imagenet_std + imagenet_mean) * 255, 0, 255).int())

    plt.title("original")

    plt.subplot(1, 4, 2)
    plt.imshow(torch.clip((im_masked[0] * imagenet_std + imagenet_mean) * 255, 0, 255).int())
    plt.title('masked')

    plt.subplot(1, 4, 3)
    plt.imshow(torch.clip((y[0] * imagenet_std + imagenet_mean) * 255, 0, 255).int())
    plt.title("reconstruction")

    plt.subplot(1, 4, 4)
    plt.imshow(torch.clip((im_paste[0] * imagenet_std + imagenet_mean) * 255, 0, 255).int())
    plt.title("reconstruction + visible")
    plt.axis('off')
    plt.savefig(output_filename)
        
img_url = 'https://user-images.githubusercontent.com/11435359/147738734-196fd92f-9260-48d5-ba7e-bf103d29364d.jpg' # fox, from ILSVRC2012_val_00046145
# img_url = 'https://user-images.githubusercontent.com/11435359/147743081-0428eecf-89e5-4e07-8da5-a30fd73cc0ba.jpg' # cucumber, from ILSVRC2012_val_00047851
img = Image.open(requests.get(img_url, stream=True).raw)
img = img.resize((224, 224))
img = np.array(img) / 255.

assert img.shape == (224, 224, 3)

# normalize by ImageNet mean and std
img = img - imagenet_mean
img = img / imagenet_std

plt.rcParams['figure.figsize'] = [5, 5]

# This is an MAE model trained with pixels as targets for visualization (ViT-Large, training mask ratio=0.75)
url = 'https://dl.fbaipublicfiles.com/mae/visualize/mae_visualize_vit_large.pth'
output_filename = '/HDD/weights/mae/mae_visualize_vit_large_75.pth'
download_weights_from_url(url, output_filename)

chkpt_dir = output_filename
model_mae = prepare_model(chkpt_dir, 'mae_vit_large_patch16')
print('Model loaded.')

# make random mask reproducible (comment out to make it change)
torch.manual_seed(2)
print('MAE with pixel reconstruction:')
run_one_image(img, model_mae, f'/HDD/etc/result.png')


url = 'https://dl.fbaipublicfiles.com/mae/visualize/mae_visualize_vit_large_ganloss.pth'
output_filename = '/HDD/weights/mae/mae_visualize_vit_large_ganloss.pth'
download_weights_from_url(url, output_filename)

chkpt_dir = output_filename
model_mae_gan = prepare_model(chkpt_dir, 'mae_vit_large_patch16')
print('Model loaded.')

torch.manual_seed(2)
print('MAE with extra GAN loss:')
run_one_image(img, model_mae_gan, f'/HDD/etc/result_ganloss.png')


import glob 
import os
import os.path as osp
import imgviz
import json
import cv2
import numpy as np
import pandas as pd
from tqdm import tqdm

from athena.src.tasks.classification.frameworks.tensorflow.models.modeling import EfficientNet


compare_mask = True
model_height = 320
model_width = 640
model_channel = 3
num_classes = 3
classes = ['rock', 'paper', 'scissors']
device = 'cuda'

weights = '/DeepLearning/etc/_athena_tests/recipes/agent/classification/tensorflow/train_unit/_outputs/CLASSIFICATION/2023_12_19_11_28_56/train/weights/last.h5'

model = EfficientNet(
            model_name='efficientnetb3',
            input_shape=(model_height, model_width, model_channel),
            num_classes=num_classes
        )

model.build(input_shape=(1, model_height, model_width, model_channel))
model.load_weights(weights)


idx2class = {ii: vv for ii, vv in enumerate(classes)}

# compare_gt = True
# iou_threshold = 0.7
# max_dets = 50
# nms_conf_threshold = 0.25
# nms_iou_threshold = 0.1
# classes = model.classes
# input_dir = '/DeepLearning/_athena_tests/datasets/directory1/split_dataset_unit/val'
# output_dir = f'/DeepLearning/etc/_athena_tests/benchmark/interojo/results/yolov7_results'

# if not osp.exists(output_dir):
#     os.makedirs(output_dir)

# img_files = glob.glob(osp.join(input_dir, '*.bmp'))

# results = {}
# compare = {}
# for img_file in tqdm(img_files):
#     filename = osp.split(osp.splitext(img_file)[0])[-1]
#     img = cv2.imread(img_file).astype(np.float32)
#     img_h, img_w, img_c = img.shape
#     img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)/255
    
#     with torch.no_grad():
#         preds = model(torch_img)
#     pred = preds[0].detach().cpu()
    
#     idx2xyxys = {}
#     for det in dets:
#         if len(det):
#             det[:, :4] = scale_coords(img.shape[:2], det[:, :4], (img_h, img_w)).round()

#             for *xyxy, conf, cls in reversed(det):
#                 cls = cls.detach().cpu().item()
#                 conf = float(conf.detach().cpu().item())
#                 if cls not in idx2xyxys.keys():
#                     idx2xyxys[cls] = []

#                 idx2xyxys[cls].append([[int(np.round(xyxy[0].detach().cpu().item())), int(np.round(xyxy[1].detach().cpu().item()))], 
#                                        [int(np.round(xyxy[2].detach().cpu().item())), int(np.round(xyxy[3].detach().cpu().item()))]])
  
#     color_map = imgviz.label_colormap()[1:len(idx2class) + 1 + 1]
#     results.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
    
#     if compare_gt:
#         _compare = vis_hbb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
#                            compare_gt=compare_gt, iou_threshold=iou_threshold)
#         _compare.update({"img_file": img_file})
#         compare.update({filename: _compare})
#     else:
#         vis_hbb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
#                 compare_gt=compare_gt, iou_threshold=iou_threshold)
            
# with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
#     json.dump(results, json_file, ensure_ascii=False, indent=4)

# if compare_gt:
#     with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
#         json.dump(compare, json_file, ensure_ascii=False, indent=4)
    
#     df_compare = pd.DataFrame(compare)
#     df_compare_pixel = df_compare.loc['diff_iou'].T
#     df_compare_pixel.fillna(0, inplace=True)
#     df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))


#!/usr/bin/env python3

import torch.nn as nn

from mmcv.cnn import ConvModule
from mmseg.ops import resize
from mmseg.models.backbones.resnet import BasicBlock


class SideConv(nn.Module):
    """'Improved' Basic Side Convolution

    instead of deconv, we use
    upsample -> 3x3conv

    https://distill.pub/2016/deconv-checkerboard/
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        conv_cfg=None,
        norm_cfg=None,
        bias=False,
        act_cfg=dict(type="ReLU"),
        interpolation="bilinear",
        align_corners=False,
    ):
        super().__init__()
        self.conv_reduce = ConvModule(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=1,
            bias=False,  # NOTE: worse results
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
        )
        self.post_resize = nn.Sequential(
            nn.ReflectionPad2d(1),
            ConvModule(
                in_channels=out_channels,
                out_channels=out_channels,
                kernel_size=3,
                stride=1,
                padding=0,
                conv_cfg=conv_cfg,
                norm_cfg=None,  # no bn
                bias=bias,
                act_cfg=None,  # no activation -> edge logit
            ),
        )
        self._interp = interpolation
        self._align_corners = align_corners

    def forward(self, x, size):
        x = resize(  # (B, out_channels, H, W)
            self.conv_reduce(x),
            size=size,
            mode=self._interp,
            align_corners=self._align_corners,
        )
        x = self.post_resize(x)
        return x


class BasicBlockSideConv(nn.Module):
    """Side Convolution with Basic Block

    Better 'information converter'
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        num_blocks=2,
        return_block_output=False,
        dilations=None,
        conv_cfg=None,
        norm_cfg=None,
        bias=False,
        act_cfg=dict(type="ReLU"),
        interpolation="bilinear",
        align_corners=False,
    ):
        super().__init__()

        assert num_blocks > 0
        self.block = self._build_block(
            num_blocks=num_blocks,
            in_channels=in_channels,
            dilations=dilations,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
        )
        self.side_conv = SideConv(
            in_channels=in_channels,
            out_channels=out_channels,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            bias=bias,
            act_cfg=act_cfg,
            interpolation=interpolation,
            align_corners=align_corners,
        )
        self.return_block_output = return_block_output

    def _build_block(
        self,
        num_blocks,
        in_channels,
        dilations,
        conv_cfg,
        norm_cfg,
    ):
        if dilations is None:
            # default is 1
            dilations = tuple([1] * num_blocks)
        assert num_blocks == len(dilations)
        modules = []
        for d in dilations:
            modules.append(
                BasicBlock(
                    inplanes=in_channels,
                    planes=in_channels,
                    stride=1,
                    dilation=d,
                    downsample=None,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                )
            )
        return nn.Sequential(*modules)

    def forward(self, x, size):
        x = self.block(x)
        s = self.side_conv(x, size)
        if self.return_block_output:
            return s, x
        else:
            return s



#!/usr/bin/env python3

from .fusion_layers import (
    GroupedConvFuse,
    GeneralizedLocationAdaptiveLearner,
)
from .side_layers import (
    BasicBlockSideConv,
    SideConv,
)


__all__ = [
    "GroupedConvFuse",
    "GeneralizedLocationAdaptiveLearner",
    "BasicBlockSideConv",
    "SideConv",
]



#!/usr/bin/env python3

"""DDS (Deep Diverse Supervision)

https://arxiv.org/pdf/1804.02864.pdf
"""

import torch.nn as nn

from .base_multisupervision_head import BaseMultiSupervisionHead
from ..builder import HEADS
from ..utils import BasicBlockSideConv, GroupedConvFuse


@HEADS.register_module()
class GeneralizedDDSHead(BaseMultiSupervisionHead):
    def __init__(
        self,
        num_blocks=2,
        dilations=None,
        side_resize_index=-1,
        edge_key="fuse",
        log_edge_keys=("fuse", "side5", "side4"),
        binary_keys=("side1", "side2", "side3", "side4"),
        multilabel_keys=("side5", "fuse"),
        loss_binary=dict(type="BinaryEdgeLoss", loss_weight=1.0),
        loss_multilabel=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            edge_key=edge_key,
            log_edge_keys=log_edge_keys,
            binary_keys=binary_keys,
            multilabel_keys=multilabel_keys,
            loss_binary=loss_binary,
            loss_multilabel=loss_multilabel,
            **kwargs,
        )

        self.side_resize_index = side_resize_index

        _interp = "bilinear"  # nearest
        _bias = True

        # bias should not be turn on when some of the sides are not supervised

        sides = []

        for i in range(len(self.in_channels) - 1):
            sides.append(
                BasicBlockSideConv(
                    in_channels=self.in_channels[i],
                    out_channels=1,
                    num_blocks=num_blocks,
                    dilations=dilations,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    bias=_bias,
                    act_cfg=self.act_cfg,
                    interpolation=_interp,
                    align_corners=self.align_corners,
                )
            )

        # last side is semantic
        sides.append(
            BasicBlockSideConv(
                in_channels=self.in_channels[-1],
                out_channels=self.num_classes,
                num_blocks=num_blocks,
                dilations=dilations,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                bias=_bias,
                act_cfg=self.act_cfg,
                interpolation=_interp,
                align_corners=self.align_corners,
            )
        )

        self.sides = nn.ModuleList(sides)
        self.fuse = GroupedConvFuse(
            num_classes=self.num_classes,
            num_sides=len(sides),
            conv_cfg=self.conv_cfg,
            bias=_bias,  # originally true
        )

    def forward(self, inputs):
        x = [i for i in inputs]
        assert isinstance(x, list)
        # get the input image size
        bs, c, h, w = x[self.side_resize_index].shape
        resize_to = (h, w)  # TODO: might be too large

        # remove the input image and unused features
        x = [x[i] for i in self.in_index]

        side_outs = []
        for i, layer in enumerate(self.sides):
            side_outs.append(layer(x[i], resize_to))

        fuse = self.fuse(side_outs)

        outs = dict(fuse=fuse)
        for i, side_out in enumerate(side_outs):
            outs[f"side{i + 1}"] = side_out

        return outs


@HEADS.register_module()
class AuxDDSHead(GeneralizedDDSHead):
    ...



from visionsuite.engines.segmentation.utils.registry import TRAINERS


def build_trainer(**config):
    trainer = TRAINERS.get(config['type'], case_sensitive=config['case_sensitive'])
    
    return trainer





from __future__ import absolute_import

from keras_unet_collection.layer_utils import *
from keras_unet_collection.activations import GELU, Snake

from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model


def RSU(X, channel_in, channel_out, depth=5, activation='ReLU', batch_norm=True, pool=True, unpool=True, name='RSU'):
    '''
    The Residual U-blocks (RSU).
    
    RSU(X, channel_in, channel_out, depth=5, activation='ReLU', batch_norm=True, pool=True, unpool=True, name='RSU')
    
    ----------
    Qin, X., Zhang, Z., Huang, C., Dehghan, M., Zaiane, O.R. and Jagersand, M., 2020. 
    U2-Net: Going deeper with nested U-structure for salient object detection. 
    Pattern Recognition, 106, p.107404.
    
    Input
    ----------
        X: input tensor.
        channel_in: number of intermediate channels.
        channel_out: number of output channels.
        depth: number of down- and upsampling levels.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        batch_norm: True for batch normalization, False otherwise.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.   
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
        
    '''
    
    pool_size = 2
    
    X_skip = []
    
    X = CONV_stack(X, channel_out, kernel_size=3, stack_num=1, 
                   dilation_rate=1, activation=activation, batch_norm=batch_norm, 
                   name='{}_in'.format(name))
    X_skip.append(X)

    X = CONV_stack(X, channel_in, kernel_size=3, stack_num=1, dilation_rate=1, 
                   activation=activation, batch_norm=batch_norm, name='{}_down_0'.format(name))
    X_skip.append(X)
    
    for i in range(depth):
        
        X = encode_layer(X, channel_in, pool_size, pool, activation=activation, 
                         batch_norm=batch_norm, name='{}_encode_{}'.format(name, i))
        
        X = CONV_stack(X, channel_in, kernel_size=3, stack_num=1, dilation_rate=1, 
                       activation=activation, batch_norm=batch_norm, name='{}_down_{}'.format(name, i+1))
        X_skip.append(X)

    X = CONV_stack(X, channel_in, kernel_size=3, stack_num=1, 
               dilation_rate=2, activation=activation, batch_norm=batch_norm, 
               name='{}_up_0'.format(name))
    
    X_skip = X_skip[::-1]
    
    for i in range(depth):
        
        X = concatenate([X, X_skip[i]], axis=-1, name='{}_concat_{}'.format(name, i))
        
        X = CONV_stack(X, channel_in, kernel_size=3, stack_num=1, dilation_rate=1, 
                       activation=activation, batch_norm=batch_norm, name='{}_up_{}'.format(name, i+1))
        
        X = decode_layer(X, channel_in, pool_size, unpool, 
                         activation=activation, batch_norm=batch_norm, name='{}_decode_{}'.format(name, i))
        
    X = concatenate([X, X_skip[depth]], axis=-1, name='{}_concat_out'.format(name))

    X = CONV_stack(X, channel_out, kernel_size=3, stack_num=1, dilation_rate=1, 
                   activation=activation, batch_norm=batch_norm, name='{}_out'.format(name))
    X = add([X, X_skip[-1]], name='{}_out_add'.format(name))
    return X 

def RSU4F(X, channel_in, channel_out, dilation_num=[1, 2, 4, 8], activation='ReLU', batch_norm=True, name='RSU4F'):
    '''
    The Residual U-blocks with dilated convolutional kernels (RSU4F).
    
    RSU4F(X, channel_in, channel_out, dilation_num=[1, 2, 4, 8], activation='ReLU', batch_norm=True, name='RSU4F')
    
    ----------
    Qin, X., Zhang, Z., Huang, C., Dehghan, M., Zaiane, O.R. and Jagersand, M., 2020. 
    U2-Net: Going deeper with nested U-structure for salient object detection. 
    Pattern Recognition, 106, p.107404.
    
    Input
    ----------
        X: input tensor.
        channel_in: number of intermediate channels.
        channel_out: number of output channels.
        dilation_num: an iterable that defines dilation rates of convolutional layers.
                      Qin et al. (2020) suggested `[1, 2, 4, 8]`.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor
        
    '''
    
    X_skip = []    
    X = CONV_stack(X, channel_out, kernel_size=3, stack_num=1, dilation_rate=1, 
                   activation=activation, batch_norm=batch_norm, name='{}_in'.format(name))
    X_skip.append(X)
    
    for i, d in enumerate(dilation_num):

        X = CONV_stack(X, channel_in, kernel_size=3, stack_num=1, dilation_rate=d, 
                       activation=activation, batch_norm=batch_norm, name='{}_down_{}'.format(name, i))
        X_skip.append(X)

    X_skip = X_skip[:-1][::-1]
    dilation_num = dilation_num[:-1][::-1]
    
    for i, d in enumerate(dilation_num[:-1]):

        X = concatenate([X, X_skip[i]], axis=-1, name='{}_concat_{}'.format(name, i))
        X = CONV_stack(X, channel_in, kernel_size=3, stack_num=1, dilation_rate=d, 
                       activation=activation, batch_norm=batch_norm, name='{}_up_{}'.format(name, i))

    X = concatenate([X, X_skip[2]], axis=-1, name='{}_concat_out'.format(name))
    X = CONV_stack(X, channel_out, kernel_size=3, stack_num=1, dilation_rate=1, 
                   activation=activation, batch_norm=batch_norm, name='{}_out'.format(name))
    
    return add([X, X_skip[-1]], name='{}_out_add'.format(name))

def u2net_2d_base(input_tensor, 
                      filter_num_down, filter_num_up, 
                      filter_mid_num_down, filter_mid_num_up, 
                      filter_4f_num, filter_4f_mid_num, activation='ReLU',
                      batch_norm=False, pool=True, unpool=True, name='u2net'):
    
    '''
    The base of U^2-Net
    
    u2net_2d_base(input_tensor, 
                  filter_num_down, filter_num_up, 
                  filter_mid_num_down, filter_mid_num_up, 
                  filter_4f_num, filter_4f_mid_num, activation='ReLU',
                  batch_norm=False, pool=True, unpool=True, name='u2net')
    
    ----------
    Qin, X., Zhang, Z., Huang, C., Dehghan, M., Zaiane, O.R. and Jagersand, M., 2020. 
    U2-Net: Going deeper with nested U-structure for salient object detection. 
    Pattern Recognition, 106, p.107404.
    
    Input
    ----------
        input_tensor: the input tensor of the base, e.g., keras.layers.Inpyt((None, None, 3))
        filter_num_down: a list that defines the number of RSU output filters for each 
                         downsampling level. e.g., `[64, 128, 256, 512]`.
                         the network depth is expected as `len(filter_num_down) + len(filter_4f_num)`                         
        filter_mid_num_down: a list that defines the number of RSU intermediate filters for each 
                             downsampling level. e.g., `[16, 32, 64, 128]`.
                             * RSU intermediate and output filters must paired, i.e., list with the same length.
                             * RSU intermediate filters numbers are expected to be smaller than output filters numbers.
        filter_mid_num_up: a list that defines the number of RSU intermediate filters for each 
                           upsampling level. e.g., `[16, 32, 64, 128]`.
                           * RSU intermediate and output filters must paired, i.e., list with the same length.
                           * RSU intermediate filters numbers are expected to be smaller than output filters numbers.
        filter_4f_num: a list that defines the number of RSU-4F output filters for each 
                       downsampling and bottom level. e.g., `[512, 512]`.
                       the network depth is expected as `len(filter_num_down) + len(filter_4f_num)`.
        filter_4f_mid_num: a list that defines the number of RSU-4F intermediate filters for each 
                           downsampling and bottom level. e.g., `[256, 256]`.
                           * RSU-4F intermediate and output filters must paired, i.e., list with the same length.
                           * RSU-4F intermediate filters numbers are expected to be smaller than output filters numbers.    
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        batch_norm: True for batch normalization.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.
        deep_supervision: True for a model that supports deep supervision. Details see Qin et al. (2020).
        name: prefix of the created keras layers.
        
    Output
    ----------
        A list of tensors with the first/second/third tensor obtained from 
        the deepest/second deepest/third deepest upsampling block, etc.
        * The feature map sizes of these tensors are different, 
          with first tensor has the smallest size.
        
    * Dilation rates of RSU4F layers are fixed to `[1, 2, 4, 8]`.
    * Downsampling is achieved through maxpooling in Qin et al. (2020), 
      and can be replaced by strided convolutional layers here.
    * Upsampling is achieved through bilinear interpolation in Qin et al. (2020), 
      and can be replaced by transpose convolutional layers here.
    
    '''
    
    pool_size = 2
    
    X_skip = []; X_out = []; OUT_stack = []
    depth_backup = []
    depth_ = len(filter_num_down)
    
    X = input_tensor
    
    X = RSU(X, filter_mid_num_down[0], filter_num_down[0], depth=depth_+1, activation=activation, 
            batch_norm=batch_norm, pool=pool, unpool=unpool, name='{}_in'.format(name))
    X_skip.append(X)
    
    depth_backup.append(depth_+1)
    
    for i, f in enumerate(filter_num_down[1:]):
        
        X = encode_layer(X, f, pool_size, pool, activation=activation, 
                         batch_norm=batch_norm, name='{}_encode_{}'.format(name, i))
        
        X = RSU(X, filter_mid_num_down[i+1], f, depth=depth_-i, activation=activation, 
                batch_norm=batch_norm, pool=pool, unpool=unpool, name='{}_down_{}'.format(name, i))
        
        depth_backup.append(depth_-i)
        
        X_skip.append(X)

    for i, f in enumerate(filter_4f_num):
        
        X = encode_layer(X, f, pool_size, pool, activation=activation, 
                         batch_norm=batch_norm, name='{}_encode_4f_{}'.format(name, i))
        
        X = RSU4F(X, filter_4f_mid_num[i], f, activation=activation, 
                  batch_norm=batch_norm, name='{}_down_4f_{}'.format(name, i))
        X_skip.append(X)
        
    X_out.append(X)
    
    # ---------- #
    X_skip = X_skip[:-1][::-1]
    depth_backup = depth_backup[::-1]
    
    filter_num_up = filter_num_up[::-1]
    filter_mid_num_up = filter_mid_num_up[::-1]
    
    filter_4f_num = filter_4f_num[:-1][::-1]
    filter_4f_mid_num = filter_4f_mid_num[:-1][::-1]
    
    tensor_count = 0
    for i, f in enumerate(filter_4f_num):
        
        X = decode_layer(X, f, pool_size, unpool, 
                         activation=activation, batch_norm=batch_norm, name='{}_decode_4f_{}'.format(name, i))
        
        X = concatenate([X, X_skip[tensor_count]], axis=-1, name='{}_concat_4f_{}'.format(name, i))
        
        X = RSU4F(X, filter_4f_mid_num[i], f, activation=activation, 
                  batch_norm=batch_norm, name='{}_up_4f_{}'.format(name, i))
        X_out.append(X)
        
        tensor_count += 1
    
    for i, f in enumerate(filter_num_up):
        
        X = decode_layer(X, f, pool_size, unpool, 
                         activation=activation, batch_norm=batch_norm, name='{}_decode_{}'.format(name, i))
        
        X = concatenate([X, X_skip[tensor_count]], axis=-1, name='{}_concat_{}'.format(name, i))
        
        X = RSU(X, filter_mid_num_up[i], f, depth=depth_backup[i], 
                activation=activation, batch_norm=batch_norm, pool=pool, unpool=unpool, name='{}_up_{}'.format(name, i))
        X_out.append(X)
        
        tensor_count += 1

    return X_out


def u2net_2d(input_size, n_labels, filter_num_down, filter_num_up='auto', filter_mid_num_down='auto', filter_mid_num_up='auto', 
             filter_4f_num='auto', filter_4f_mid_num='auto', activation='ReLU', output_activation='Sigmoid', 
             batch_norm=False, pool=True, unpool=True, deep_supervision=False, name='u2net'):
    
    '''
    U^2-Net
    
    u2net_2d(input_size, n_labels, filter_num_down, filter_num_up='auto', filter_mid_num_down='auto', filter_mid_num_up='auto', 
             filter_4f_num='auto', filter_4f_mid_num='auto', activation='ReLU', output_activation='Sigmoid', 
             batch_norm=False, deep_supervision=False, name='u2net')
    
    ----------
    Qin, X., Zhang, Z., Huang, C., Dehghan, M., Zaiane, O.R. and Jagersand, M., 2020. 
    U2-Net: Going deeper with nested U-structure for salient object detection. 
    Pattern Recognition, 106, p.107404.
    
    Input
    ----------
        input_size: the size/shape of network input, e.g., `(128, 128, 3)`.
        filter_num_down: a list that defines the number of RSU output filters for each 
                         downsampling level. e.g., `[64, 128, 256, 512]`.
                         the network depth is expected as `len(filter_num_down) + len(filter_4f_num)`                         
        filter_mid_num_down: a list that defines the number of RSU intermediate filters for each 
                             downsampling level. e.g., `[16, 32, 64, 128]`.
                             * RSU intermediate and output filters must paired, i.e., list with the same length.
                             * RSU intermediate filters numbers are expected to be smaller than output filters numbers.
        filter_mid_num_up: a list that defines the number of RSU intermediate filters for each 
                           upsampling level. e.g., `[16, 32, 64, 128]`.
                           * RSU intermediate and output filters must paired, i.e., list with the same length.
                           * RSU intermediate filters numbers are expected to be smaller than output filters numbers.
        filter_4f_num: a list that defines the number of RSU-4F output filters for each 
                       downsampling and bottom level. e.g., `[512, 512]`.
                       the network depth is expected as `len(filter_num_down) + len(filter_4f_num)`.
        filter_4f_mid_num: a list that defines the number of RSU-4F intermediate filters for each 
                           downsampling and bottom level. e.g., `[256, 256]`.
                           * RSU-4F intermediate and output filters must paired, i.e., list with the same length.
                           * RSU-4F intermediate filters numbers are expected to be smaller than output filters numbers.         
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        output_activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interface or 'Sigmoid'.
                           Default option is 'Softmax'.
                           if None is received, then linear activation is applied.           
        batch_norm: True for batch normalization.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.  
        deep_supervision: True for a model that supports deep supervision. Details see Qin et al. (2020).
        name: prefix of the created keras layers.
        
    Output
    ----------
        model: a keras model.
    
    * Automated hyper-parameter estimation will produce a slightly larger network, different from that of Qin et al. (2020).
    * Dilation rates of RSU4F layers are fixed to `[1, 2, 4, 8]`.
    * The default output activation is sigmoid, the same as in Qin et al. (2020).
    * Downsampling is achieved through maxpooling and can be replaced by strided convolutional layers.
    * Upsampling is achieved through bilinear interpolation and can be replaced by transpose convolutional layers.
    
    '''
    
    verbose = False
    
    if filter_num_up == 'auto':
        verbose = True
        filter_num_up = filter_num_down
        
    if filter_mid_num_down == 'auto':
        verbose = True
        filter_mid_num_down = [num//4 for num in filter_num_down]
        
    if filter_mid_num_up == 'auto':
        verbose = True
        filter_mid_num_up = filter_mid_num_down
        
    if filter_4f_num == 'auto':
        verbose = True
        filter_4f_num = [filter_num_down[-1], filter_num_down[-1]]
        
    if filter_4f_mid_num == 'auto':
        verbose = True
        filter_4f_mid_num = [num//2 for num in filter_4f_num]
        
    if verbose:
        print('Automated hyper-parameter determination is applied with the following details:\n----------')
        print('\tNumber of RSU output channels within downsampling blocks: filter_num_down = {}'.format(filter_num_down))
        print('\tNumber of RSU intermediate channels within downsampling blocks: filter_mid_num_down = {}'.format(filter_mid_num_down))
        print('\tNumber of RSU output channels within upsampling blocks: filter_num_up = {}'.format(filter_num_up))
        print('\tNumber of RSU intermediate channels within upsampling blocks: filter_mid_num_up = {}'.format(filter_mid_num_up))        
        print('\tNumber of RSU-4F output channels within downsampling and bottom blocks: filter_4f_num = {}'.format(filter_4f_num))
        print('\tNumber of RSU-4F intermediate channels within downsampling and bottom blocks: filter_4f_num = {}'.format(filter_4f_mid_num))
        print('----------\nExplicitly specifying keywords listed above if their "auto" settings do not satisfy your needs')
        
    print("----------\nThe depth of u2net_2d = len(filter_num_down) + len(filter_4f_num) = {}".format(len(filter_num_down)+len(filter_4f_num)))
    
    X_skip = []; X_out = []; OUT_stack = []
    depth_backup = []
    depth_ = len(filter_num_down)
    
    IN = Input(shape=input_size) 
    
    # base (before conv + activation + upsample)
    X_out = u2net_2d_base(IN, 
                          filter_num_down, filter_num_up, 
                          filter_mid_num_down, filter_mid_num_up, 
                          filter_4f_num, filter_4f_mid_num, activation=activation, 
                          batch_norm=batch_norm, pool=pool, unpool=unpool, name=name)
    
    # output layers
    X_out = X_out[::-1]
    L_out = len(X_out)
    
    X = CONV_output(X_out[0], n_labels, kernel_size=3, activation=output_activation, 
                    name='{}_output_sup0'.format(name))
    OUT_stack.append(X)
    
    for i in range(1, L_out):
        
        pool_size = 2**(i)
        
        X = Conv2D(n_labels, 3, padding='same', name='{}_output_conv_{}'.format(name, i))(X_out[i])
        
        X = decode_layer(X, n_labels, pool_size, unpool, 
                         activation=None, batch_norm=False, name='{}_sup{}'.format(name, i))
        
        if output_activation:
            if output_activation == 'Sigmoid':
                X = Activation('sigmoid', name='{}_output_sup{}_activation'.format(name, i))(X)
            else:
                activation_func = eval(output_activation)
                X = activation_func(name='{}_output_sup{}_activation'.format(name, i))(X)
                
        OUT_stack.append(X)
        
    D = concatenate(OUT_stack, axis=-1, name='{}_output_concat'.format(name))
    
    D = CONV_output(D, n_labels, kernel_size=1, activation=output_activation, 
                    name='{}_output_final'.format(name))
    
    if deep_supervision:
        
        OUT_stack.append(D)
        print('----------\ndeep_supervision = True\nnames of output tensors are listed as follows ("sup0" is the shallowest supervision layer;\n"final" is the final output layer):\n')
        
        if output_activation == None:
            if unpool is False:
                for i in range(L_out):
                    print('\t{}_output_sup{}_trans_conv'.format(name, i))
            else:
                for i in range(L_out):
                    print('\t{}_output_sup{}_unpool'.format(name, i))
                
            print('\t{}_output_final'.format(name))
        
        else:        
            for i in range(L_out):
                print('\t{}_output_sup{}_activation'.format(name, i))
                
            print('\t{}_output_final_activation'.format(name))
            
        model = Model([IN,], OUT_stack)
        
    else:
        model = Model([IN,], [D,])
        
    return model



import imgviz
import numpy as np
import cv2
import os.path as osp
import os 
import torch

def vis_by_batch(output, target, image, fname, epoch, batch_idx, output_dir, denormalize=None,
              input_channel=3, origin = (25,25), font = cv2.FONT_HERSHEY_SIMPLEX):
    
    for step_idx, (pred, mask, image, fname) in enumerate(zip(output, target, image, fname)):
        vis_by_step(pred, mask, image, fname, epoch, batch_idx, step_idx, output_dir, denormalize=denormalize, 
                    input_channel=input_channel, origin=origin)
        
                
def vis_by_step(pred, mask, image, fname, epoch, batch_idx, step_idx, output_dir, denormalize=None, 
                input_channel=3, origin = (25,25), font = cv2.FONT_HERSHEY_SIMPLEX):
    pred = torch.nn.functional.softmax(pred, dim=0)
    pred = torch.argmax(pred, dim=0)
    pred = pred.detach().float().to('cpu')
    pred = pred.numpy()
    
    image = image.to('cpu')
    image = image.numpy()
    image = image.transpose((1, 2, 0))
    image = denormalize(image)
    color_map = imgviz.label_colormap(256)
    mask = mask.cpu().detach()
    mask = color_map[mask.numpy().astype(np.uint8)].astype(np.uint8)
    pred = color_map[pred.astype(np.uint8)].astype(np.uint8)
    
    pred = cv2.addWeighted(image, 0.1, pred, 0.9, 0)
    mask = cv2.addWeighted(image, 0.1, mask, 0.9, 0)

    text1 = np.zeros((50, image.shape[1], input_channel), np.uint8)
    text2 = np.zeros((50, image.shape[1], input_channel), np.uint8)
    text3 = np.zeros((50, image.shape[1], input_channel), np.uint8)
    cv2.putText(text1, "(a) original", origin, font, 0.6, (255,255,255), 1)
    cv2.putText(text2, "(b) ground truth" , origin, font, 0.6, (255,255,255), 1)
    cv2.putText(text3, "(c) predicted" , origin, font, 0.6, (255,255,255), 1)

    image = cv2.vconcat([text1, image])
    mask = cv2.vconcat([text2, mask])
    pred = cv2.vconcat([text3, pred.astype(np.uint8)])

    res = cv2.hconcat([image, mask, pred])
    vis_dir = osp.join(output_dir, str(epoch))
    if not osp.exists(vis_dir):
        os.mkdir(vis_dir)
    cv2.imwrite(osp.join(vis_dir, f"{fname}_{batch_idx}_{step_idx}.png"), res)


# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A set of functions that are used for visualization.

These functions often receive an image, perform some visualization on the image.
The functions do not return a value, instead they modify the image itself.

"""
import collections
import numpy as np
import PIL.Image as Image
import PIL.ImageColor as ImageColor
import PIL.ImageDraw as ImageDraw
import PIL.ImageFont as ImageFont
import matplotlib.pyplot as plt
_TITLE_LEFT_MARGIN = 10
_TITLE_TOP_MARGIN = 10
STANDARD_COLORS = [
  'AliceBlue', 'Chartreuse', 'Aqua', 'Aquamarine', 'Azure', 'Beige', 'Bisque',
  'BlanchedAlmond', 'BlueViolet', 'BurlyWood', 'CadetBlue', 'AntiqueWhite',
  'Chocolate', 'Coral', 'CornflowerBlue', 'Cornsilk', 'Crimson', 'Cyan',
  'DarkCyan', 'DarkGoldenRod', 'DarkGrey', 'DarkKhaki', 'DarkOrange',
  'DarkOrchid', 'DarkSalmon', 'DarkSeaGreen', 'DarkTurquoise', 'DarkViolet',
  'DeepPink', 'DeepSkyBlue', 'DodgerBlue', 'FireBrick', 'FloralWhite',
  'ForestGreen', 'Fuchsia', 'Gainsboro', 'GhostWhite', 'Gold', 'GoldenRod',
  'Salmon', 'Tan', 'HoneyDew', 'HotPink', 'IndianRed', 'Ivory', 'Khaki',
  'Lavender', 'LavenderBlush', 'LawnGreen', 'LemonChiffon', 'LightBlue',
  'LightCoral', 'LightCyan', 'LightGoldenRodYellow', 'LightGray', 'LightGrey',
  'LightGreen', 'LightPink', 'LightSalmon', 'LightSeaGreen', 'LightSkyBlue',
  'LightSlateGray', 'LightSlateGrey', 'LightSteelBlue', 'LightYellow', 'Lime',
  'LimeGreen', 'Linen', 'Magenta', 'MediumAquaMarine', 'MediumOrchid',
  'MediumPurple', 'MediumSeaGreen', 'MediumSlateBlue', 'MediumSpringGreen',
  'MediumTurquoise', 'MediumVioletRed', 'MintCream', 'MistyRose', 'Moccasin',
  'NavajoWhite', 'OldLace', 'Olive', 'OliveDrab', 'Orange', 'OrangeRed',
  'Orchid', 'PaleGoldenRod', 'PaleGreen', 'PaleTurquoise', 'PaleVioletRed',
  'PapayaWhip', 'PeachPuff', 'Peru', 'Pink', 'Plum', 'PowderBlue', 'Purple',
  'Red', 'RosyBrown', 'RoyalBlue', 'SaddleBrown', 'Green', 'SandyBrown',
  'SeaGreen', 'SeaShell', 'Sienna', 'Silver', 'SkyBlue', 'SlateBlue',
  'SlateGray', 'SlateGrey', 'Snow', 'SpringGreen', 'SteelBlue', 'GreenYellow',
  'Teal', 'Thistle', 'Tomato', 'Turquoise', 'Violet', 'Wheat', 'White',
  'WhiteSmoke', 'Yellow', 'YellowGreen'
]

def visualize_boxes(
    #    image, boxes, labels, probs, class_labels):

    #  category_index = {}
    #  for id_, label_name in enumerate(class_labels):
    #    category_index[id_] = {"name": label_name}
    #  image=visualize_boxes_and_labels_on_image_array(image, boxes, labels, probs, category_index)
    #  return image

    #def visualize_boxes_and_labels_on_image_array(
    image,
    boxes,
    labels, # classes,
    scores,
    class_labels, # category_index,
    palette=None,                                 # New!
    pred_box_visualization_color=None,            # New!
    instance_masks=None,
    instance_boundaries=None,
    use_normalized_coordinates=False,
    max_boxes_to_draw=20,
    min_score_thresh=.5,
    agnostic_mode=False,
    line_thickness=4,
    groundtruth_box_visualization_color='black',
    skip_scores=False,
    skip_labels=False):
  """I have forgot where this file came from. If you know, please tell me.

  Now `boxes` supports input:
    1. xmin, ymin, xmax, ymax
    2. x_ctr, y_ctr, w, h, angle
    3. x0, y0, x1, y1, x2, y2, x3, y3

  Overlay labeled boxes on an image with formatted scores and label names.

  This function groups boxes that correspond to the same location
  and creates a display string for each detection and overlays these
  on the image. Note that this function modifies the image in place, and returns
  that same image.

  Args:
      image: uint8 numpy array with shape (img_height, img_width, 3)
      boxes: a numpy array of shape [N, 4]
      classes: a numpy array of shape [N]. Note that class indices are 1-based,
          and match the keys in the label map.
      scores: a numpy array of shape [N] or None.    If scores=None, then
          this function assumes that the boxes to be plotted are groundtruth
          boxes and plot all boxes as black with no classes or scores.
      category_index: a dict containing category dictionaries (each holding
          category index `id` and category name `name`) keyed by category indices.
      instance_masks: a numpy array of shape [N, image_height, image_width] with
          values ranging between 0 and 1, can be None.
      instance_boundaries: a numpy array of shape [N, image_height, image_width]
          with values ranging between 0 and 1, can be None.
      use_normalized_coordinates: whether boxes is to be interpreted as
          normalized coordinates or not.
      max_boxes_to_draw: maximum number of boxes to visualize.    If None, draw
          all boxes.
      min_score_thresh: minimum score threshold for a box to be visualized
      agnostic_mode: boolean (default: False) controlling whether to evaluate in
          class-agnostic mode or not.    This mode will display scores but ignore
          classes.
      line_thickness: integer (default: 4) controlling line width of the boxes.
      groundtruth_box_visualization_color: box color for visualizing groundtruth
          boxes
      skip_scores: whether to skip score when drawing a single detection
      skip_labels: whether to skip label when drawing a single detection

  Returns:
      uint8 numpy array with shape (img_height, img_width, 3) with overlaid boxes.
  """
  # Create a display string (and color) for every box location, group any boxes
  # that correspond to the same location.

  classes = labels

  category_index = {}
  for id_, label_name in enumerate(class_labels):
    category_index[id_] = {"name": label_name}

  box_to_display_str_map = collections.defaultdict(list)
  box_to_color_map = collections.defaultdict(str)
  box_to_instance_masks_map = {}
  box_to_instance_boundaries_map = {}
  if not max_boxes_to_draw:
    max_boxes_to_draw = boxes.shape[0]

  sorted_ind = np.argsort(-scores)
  boxes=boxes[sorted_ind]
  scores=scores[sorted_ind]
  classes=classes[sorted_ind]
  for i in range(min(max_boxes_to_draw, boxes.shape[0])):
    if scores is None or scores[i] > min_score_thresh:
      box = tuple(boxes[i].tolist())
      if instance_masks is not None:
        box_to_instance_masks_map[box] = instance_masks[i]
      if instance_boundaries is not None:
        box_to_instance_boundaries_map[box] = instance_boundaries[i]
      if scores is None:
        box_to_color_map[box] = groundtruth_box_visualization_color
      else:
        display_str = ''
        if not skip_labels:
          if not agnostic_mode:
            if classes[i] in category_index.keys():
              class_name = category_index[classes[i]]['name']
            else:
              class_name = 'N/A'
            display_str = str(class_name)
        if not skip_scores:
          if not display_str:
            display_str = '{}%'.format(int(100 * scores[i]))
          else:
            display_str = '{}: {}%'.format(display_str, int(100 * scores[i]))
        box_to_display_str_map[box].append(display_str)
        if agnostic_mode:
          box_to_color_map[box] = 'DarkOrange'
        elif pred_box_visualization_color is not None:
          box_to_color_map[box] = pred_box_visualization_color
        elif palette is not None:
          box_to_color_map[box] = palette[classes[i]]
        else:
          box_to_color_map[box] = STANDARD_COLORS[classes[i] % len(STANDARD_COLORS)]

  # Draw all boxes onto image.
  for box, color in box_to_color_map.items():
    if len(box) == 4:
        xmin, ymin, xmax, ymax = box
        if instance_masks is not None:
            draw_mask_on_image_array(
            image,
            box_to_instance_masks_map[box],
            color=color
            )
        if instance_boundaries is not None:
            draw_mask_on_image_array(
            image,
            box_to_instance_boundaries_map[box],
            color='red',
            alpha=1.0
            )
        draw_bounding_box_on_image_array(
            image,
            ymin,
            xmin,
            ymax,
            xmax,
            color=color,
            thickness=line_thickness,
            display_str_list=box_to_display_str_map[box],
            use_normalized_coordinates=use_normalized_coordinates)
    elif len(box) == 5:
        x0, y0, x1, y1, x2, y2, x3, y3 = obb2poly_np(box)
        if instance_masks is not None:
            draw_mask_on_image_array(
            image,
            box_to_instance_masks_map[box],
            color=color
            )
        if instance_boundaries is not None:
            draw_mask_on_image_array(
            image,
            box_to_instance_boundaries_map[box],
            color='red',
            alpha=1.0
            )
        draw_bounding_box_on_image_array_poly(
            image,
            x0, y0, x1, y1, x2, y2, x3, y3,
            color=color,
            thickness=line_thickness,
            display_str_list=box_to_display_str_map[box],
            use_normalized_coordinates=use_normalized_coordinates)
    elif len(box) == 8:
        x0, y0, x1, y1, x2, y2, x3, y3 = box
        if instance_masks is not None:
            draw_mask_on_image_array(
            image,
            box_to_instance_masks_map[box],
            color=color
            )
        if instance_boundaries is not None:
            draw_mask_on_image_array(
            image,
            box_to_instance_boundaries_map[box],
            color='red',
            alpha=1.0
            )
        draw_bounding_box_on_image_array_poly(
            image,
            x0, y0, x1, y1, x2, y2, x3, y3,
            color=color,
            thickness=line_thickness,
            display_str_list=box_to_display_str_map[box],
            use_normalized_coordinates=use_normalized_coordinates)
    else:
        raise ValueError(f"Invalid length of box")
  return image


def draw_bounding_box_on_image_array(image,
                                     ymin,
                                     xmin,
                                     ymax,
                                     xmax,
                                     color='red',
                                     thickness=4,
                                     display_str_list=(),
                                     use_normalized_coordinates=True):
  """Adds a bounding box to an image (numpy array).

  Bounding box coordinates can be specified in either absolute (pixel) or
  normalized coordinates by setting the use_normalized_coordinates argument.

  Args:
      image: a numpy array with shape [height, width, 3].
      ymin: ymin of bounding box.
      xmin: xmin of bounding box.
      ymax: ymax of bounding box.
      xmax: xmax of bounding box.
      color: color to draw bounding box. Default is red.
      thickness: line thickness. Default value is 4.
      display_str_list: list of strings to display in box
                                          (each to be shown on its own line).
      use_normalized_coordinates: If True (default), treat coordinates
          ymin, xmin, ymax, xmax as relative to the image.    Otherwise treat
          coordinates as absolute.
  """
  image_pil = Image.fromarray(np.uint8(image)).convert('RGB')
  draw_bounding_box_on_image(image_pil, ymin, xmin, ymax, xmax, color,
                             thickness, display_str_list,
                             use_normalized_coordinates)
  np.copyto(image, np.array(image_pil))


def draw_bounding_box_on_image(image,
                               ymin,
                               xmin,
                               ymax,
                               xmax,
                               color='red',
                               thickness=4,
                               display_str_list=(),
                               use_normalized_coordinates=True):
  """Adds a bounding box to an image.

  Bounding box coordinates can be specified in either absolute (pixel) or
  normalized coordinates by setting the use_normalized_coordinates argument.

  Each string in display_str_list is displayed on a separate line above the
  bounding box in black text on a rectangle filled with the input 'color'.
  If the top of the bounding box extends to the edge of the image, the strings
  are displayed below the bounding box.

  Args:
      image: a PIL.Image object.
      ymin: ymin of bounding box.
      xmin: xmin of bounding box.
      ymax: ymax of bounding box.
      xmax: xmax of bounding box.
      color: color to draw bounding box. Default is red.
      thickness: line thickness. Default value is 4.
      display_str_list: list of strings to display in box
                                          (each to be shown on its own line).
      use_normalized_coordinates: If True (default), treat coordinates
          ymin, xmin, ymax, xmax as relative to the image.    Otherwise treat
          coordinates as absolute.
  """
  draw = ImageDraw.Draw(image)
  im_width, im_height = image.size
  if use_normalized_coordinates:
    (left, right, top, bottom) = (xmin * im_width, xmax * im_width,
                                  ymin * im_height, ymax * im_height)
  else:
    (left, right, top, bottom) = (xmin, xmax, ymin, ymax)
  draw.line([(left, top), (left, bottom), (right, bottom),
             (right, top), (left, top)], width=thickness, fill=color)
  try:
    font = ImageFont.truetype('arial.ttf', 24)
  except IOError:
    font = ImageFont.load_default()

  # If the total height of the display strings added to the top of the bounding
  # box exceeds the top of the image, stack the strings below the bounding box
  # instead of above.
  display_str_heights = [font.getsize(ds)[1] for ds in display_str_list]
  # Each display_str has a top and bottom margin of 0.05x.
  total_display_str_height = (1 + 2 * 0.05) * sum(display_str_heights)

  if top > total_display_str_height:
    text_bottom = top
  else:
    text_bottom = bottom + total_display_str_height
  # Reverse list and print from bottom to top.
  for display_str in display_str_list[::-1]:
    text_width, text_height = font.getsize(display_str)
    margin = np.ceil(0.05 * text_height)
    draw.rectangle(
      [(left, text_bottom - text_height - 2 * margin), (left + text_width,
                                                        text_bottom)],
      fill=color)
    draw.text(
      (left + margin, text_bottom - text_height - margin),
      display_str,
      fill='black',
      font=font)
    text_bottom -= text_height - 2 * margin


def draw_mask_on_image_array(image, mask, color='red', alpha=0.4):
  """Draws mask on an image.

  Args:
      image: uint8 numpy array with shape (img_height, img_height, 3)
      mask: a uint8 numpy array of shape (img_height, img_height) with
          values between either 0 or 1.
      color: color to draw the keypoints with. Default is red.
      alpha: transparency value between 0 and 1. (default: 0.4)

  Raises:
      ValueError: On incorrect data type for image or masks.
  """
  if image.dtype != np.uint8:
    raise ValueError('`image` not of type np.uint8')
  if mask.dtype != np.uint8:
    raise ValueError('`mask` not of type np.uint8')
  if np.any(np.logical_and(mask != 1, mask != 0)):
    raise ValueError('`mask` elements should be in [0, 1]')
  if image.shape[:2] != mask.shape:
    raise ValueError('The image has spatial dimensions %s but the mask has '
                     'dimensions %s' % (image.shape[:2], mask.shape))
  rgb = ImageColor.getrgb(color)
  pil_image = Image.fromarray(image)

  solid_color = np.expand_dims(
    np.ones_like(mask), axis=2) * np.reshape(list(rgb), [1, 1, 3])
  pil_solid_color = Image.fromarray(np.uint8(solid_color)).convert('RGBA')
  pil_mask = Image.fromarray(np.uint8(255.0 * alpha * mask)).convert('L')
  pil_image = Image.composite(pil_solid_color, pil_image, pil_mask)
  np.copyto(image, np.array(pil_image.convert('RGB')))


# ===== ===== ====== ===== ===== ===== ===== ====== ===== ===== 


def draw_bounding_box_on_image_array_poly(image,
                                     x0, y0, x1, y1, x2, y2, x3, y3,
                                     color='red',
                                     thickness=4,
                                     display_str_list=(),
                                     use_normalized_coordinates=True):
  """from def draw_bounding_box_on_image_array()
  """
  image_pil = Image.fromarray(np.uint8(image)).convert('RGB')
  draw_bounding_box_on_image_poly(image_pil, x0, y0, x1, y1, x2, y2, x3, y3, color,
                             thickness, display_str_list,
                             use_normalized_coordinates)
  np.copyto(image, np.array(image_pil))


def draw_bounding_box_on_image_poly(image,
                               x0, y0, x1, y1, x2, y2, x3, y3,
                               color='red',
                               thickness=4,
                               display_str_list=(),
                               use_normalized_coordinates=True):
  """from def draw_bounding_box_on_image()
  """
  draw = ImageDraw.Draw(image)
  im_width, im_height = image.size

  # ##### ##### ##### ##### ##### 
  if use_normalized_coordinates:
    (pt1_x, pt1_y, pt2_x, pt2_y, pt3_x, pt3_y, pt4_x, pt4_y) = (x0 * im_width, y0 * im_height, 
                                                                x1 * im_width, y1 * im_height, 
                                                                x2 * im_width, y2 * im_height, 
                                                                x3 * im_width, y3 * im_height)
  else:
    (pt1_x, pt1_y, pt2_x, pt2_y, pt3_x, pt3_y, pt4_x, pt4_y) = (x0, y0, x1, y1, x2, y2, x3, y3)
  draw.line([(pt1_x, pt1_y), (pt2_x, pt2_y), (pt3_x, pt3_y),
             (pt4_x, pt4_y), (pt1_x, pt1_y)], width=thickness, fill=color)
  # ##### ##### ##### ##### ##### 

  try:
    font = ImageFont.truetype('arial.ttf', 24)
  except IOError:
    font = ImageFont.load_default()

  # If the total height of the display strings added to the top of the bounding
  # box exceeds the top of the image, stack the strings below the bounding box
  # instead of above.
  display_str_heights = [font.getsize(ds)[1] for ds in display_str_list]
  # Each display_str has a top and bottom margin of 0.05x.
  total_display_str_height = (1 + 2 * 0.05) * sum(display_str_heights)

  top = min((pt1_y, pt2_y, pt3_y, pt4_y))
  bottom = max((pt1_y, pt2_y, pt3_y, pt4_y))
  if   top == pt1_y: left = pt1_x
  elif top == pt2_y: left = pt2_x
  elif top == pt3_y: left = pt3_x
  elif top == pt4_y: left = pt4_x
  else: pass

  if top > total_display_str_height:
    text_bottom = top
  else:
    text_bottom = bottom + total_display_str_height
  # Reverse list and print from bottom to top.
  for display_str in display_str_list[::-1]:
    text_width, text_height = font.getsize(display_str)
    margin = np.ceil(0.05 * text_height)
    draw.rectangle(
      [(left, text_bottom - text_height - 2 * margin), (left + text_width,
                                                        text_bottom)],
      fill=color)
    draw.text(
      (left + margin, text_bottom - text_height - margin),
      display_str,
      fill='black',
      font=font)
    text_bottom -= text_height - 2 * margin


# ===== ===== ====== ===== ===== ===== ===== ====== ===== ===== 


def obb2poly_np(obboxes):
    """Convert oriented bounding boxes to polygons.

    Args:
        obbs (ndarray): [x_ctr,y_ctr,w,h,angle]

    Returns:
        polys (ndarray): [x0,y0,x1,y1,x2,y2,x3,y3]
    """
    try:
        center, w, h, theta = np.split(obboxes, (2, 3, 4), axis=-1)
    except:  # noqa: E722
        results = np.stack([0., 0., 0., 0., 0., 0., 0., 0.], axis=-1)
        return results.reshape(1, -1)
    Cos, Sin = np.cos(theta), np.sin(theta)
    vector1 = np.concatenate([w / 2 * Cos, w / 2 * Sin], axis=-1)
    vector2 = np.concatenate([-h / 2 * Sin, h / 2 * Cos], axis=-1)
    point1 = center - vector1 - vector2
    point2 = center + vector1 - vector2
    point3 = center + vector1 + vector2
    point4 = center - vector1 + vector2
    polys = np.concatenate([point1, point2, point3, point4], axis=-1)
    # polys = get_best_begin_point(polys)
    return polys



import inspect
import logging
from typing import TYPE_CHECKING, Any, Optional, Union

# from mmengine.config import Config, ConfigDict
from .manager import ManagerMixin
from .registry import Registry

# if TYPE_CHECKING:
#     import torch.nn as nn

#     from mmengine.optim.scheduler import _ParamScheduler
#     from mmengine.runner import Runner


def build_from_cfg(
        # cfg: Union[dict, ConfigDict, Config],
        cfg: Union[dict],
        registry: Registry,
        default_args: Optional[Union[dict]] = None) -> Any:
    # default_args: Optional[Union[dict, ConfigDict, Config]] = None) -> Any:
    """
    Build a module from config dict when it is a class configuration, or call a function from config dict when it is a
    function configuration.

    If the global variable default scope (:obj:`DefaultScope`) exists,
    :meth:`build` will firstly get the responding registry and then call
    its own :meth:`build`.

    At least one of the ``cfg`` and ``default_args`` contains the key "type",
    which should be either str or class. If they all contain it, the key
    in ``cfg`` will be used because ``cfg`` has a high priority than
    ``default_args`` that means if a key exists in both of them, the value of
    the key will be ``cfg[key]``. They will be merged first and the key "type"
    will be popped up and the remaining keys will be used as initialization
    arguments.

    Examples:
        >>> from mmengine import Registry, build_from_cfg
        >>> MODELS = Registry('models')
        >>> @MODELS.register_module()
        >>> class ResNet:
        >>>     def __init__(self, depth, stages=4):
        >>>         self.depth = depth
        >>>         self.stages = stages
        >>> cfg = dict(type='ResNet', depth=50)
        >>> model = build_from_cfg(cfg, MODELS)
        >>> # Returns an instantiated object
        >>> @MODELS.register_module()
        >>> def resnet50():
        >>>     pass
        >>> resnet = build_from_cfg(dict(type='resnet50'), MODELS)
        >>> # Return a result of the calling function

    Args:
        cfg (dict or ConfigDict or Config): Config dict. It should at least
            contain the key "type".
        registry (:obj:`Registry`): The registry to search the type from.
        default_args (dict or ConfigDict or Config, optional): Default
            initialization arguments. Defaults to None.

    Returns:
        object: The constructed object.
    """
    # Avoid circular import

    if not isinstance(cfg, (dict)):
        # if not isinstance(cfg, (dict, ConfigDict, Config)):
        raise TypeError(f'cfg should be a dict, ConfigDict or Config, but got {type(cfg)}')

    if 'type' not in cfg:
        if default_args is None or 'type' not in default_args:
            raise KeyError('`cfg` or `default_args` must contain the key "type", '
                           f'but got {cfg}\n{default_args}')

    if not isinstance(registry, Registry):
        raise TypeError('registry must be a mmengine.Registry object, '
                        f'but got {type(registry)}')

    if not (isinstance(
            default_args,
            #    (dict, ConfigDict, Config)) or default_args is None):
        (dict)) or default_args is None):
        raise TypeError('default_args should be a dict, ConfigDict, Config or None, '
                        f'but got {type(default_args)}')

    args = cfg.copy()
    if default_args is not None:
        for name, value in default_args.items():
            args.setdefault(name, value)

    # Instance should be built under target scope, if `_scope_` is defined
    # in cfg, current default scope should switch to specified scope
    # temporarily.
    scope = args.pop('_scope_', None)
    with registry.switch_scope_and_registry(scope) as registry:
        obj_type = args.pop('type')
        if isinstance(obj_type, str):
            obj_cls = registry.get(obj_type)
            if obj_cls is None:
                raise KeyError(
                    f'{obj_type} is not in the {registry.scope}::{registry.name} registry. '  # noqa: E501
                    f'Please check whether the value of `{obj_type}` is '
                    'correct or it was registered as expected. More details '
                    'can be found at '
                    'https://mmengine.readthedocs.io/en/latest/advanced_tutorials/config.html#import-the-custom-module'  # noqa: E501
                )
        # this will include classes, functions, partial functions and more
        elif callable(obj_type):
            obj_cls = obj_type
        else:
            raise TypeError(f'type must be a str or valid type, but got {type(obj_type)}')

        # If `obj_cls` inherits from `ManagerMixin`, it should be
        # instantiated by `ManagerMixin.get_instance` to ensure that it
        # can be accessed globally.
        if inspect.isclass(obj_cls) and \
                issubclass(obj_cls, ManagerMixin):  # type: ignore
            obj = obj_cls.get_instance(**args)  # type: ignore
        else:
            obj = obj_cls(**args)  # type: ignore

        if (inspect.isclass(obj_cls) or inspect.isfunction(obj_cls) or inspect.ismethod(obj_cls)):
            print(
                f'An `{obj_cls.__name__}` instance is built from '  # type: ignore # noqa: E501
                'registry, and its implementation can be found in '
                f'{obj_cls.__module__}',  # type: ignore
            )
        else:
            print('An instance is built from registry, and its constructor '
                  f'is {obj_cls}', )
        return obj


# def build_runner_from_cfg(cfg: Union[dict, ConfigDict, Config],
#                           registry: Registry) -> 'Runner':
#     """Build a Runner object.

#     Examples:
#         >>> from mmengine.registry import Registry, build_runner_from_cfg
#         >>> RUNNERS = Registry('runners', build_func=build_runner_from_cfg)
#         >>> @RUNNERS.register_module()
#         >>> class CustomRunner(Runner):
#         >>>     def setup_env(env_cfg):
#         >>>         pass
#         >>> cfg = dict(runner_type='CustomRunner', ...)
#         >>> custom_runner = RUNNERS.build(cfg)

#     Args:
#         cfg (dict or ConfigDict or Config): Config dict. If "runner_type" key
#             exists, it will be used to build a custom runner. Otherwise, it
#             will be used to build a default runner.
#         registry (:obj:`Registry`): The registry to search the type from.

#     Returns:
#         object: The constructed runner object.
#     """
#     from ..config import Config, ConfigDict
#     from ..logging import print_log

#     assert isinstance(
#         cfg,
#         (dict, ConfigDict, Config
#          )), f'cfg should be a dict, ConfigDict or Config, but got {type(cfg)}'
#     assert isinstance(
#         registry, Registry), ('registry should be a mmengine.Registry object',
#                               f'but got {type(registry)}')

#     args = cfg.copy()
#     # Runner should be built under target scope, if `_scope_` is defined
#     # in cfg, current default scope should switch to specified scope
#     # temporarily.
#     scope = args.pop('_scope_', None)
#     with registry.switch_scope_and_registry(scope) as registry:
#         obj_type = args.get('runner_type', 'Runner')
#         if isinstance(obj_type, str):
#             runner_cls = registry.get(obj_type)
#             if runner_cls is None:
#                 raise KeyError(
#                     f'{obj_type} is not in the {registry.name} registry. '
#                     f'Please check whether the value of `{obj_type}` is '
#                     'correct or it was registered as expected. More details '
#                     'can be found at https://mmengine.readthedocs.io/en/latest/advanced_tutorials/config.html#import-the-custom-module'  # noqa: E501
#                 )
#         elif inspect.isclass(obj_type):
#             runner_cls = obj_type
#         else:
#             raise TypeError(
#                 f'type must be a str or valid type, but got {type(obj_type)}')

#         runner = runner_cls.from_cfg(args)  # type: ignore
#         print_log(
#             f'An `{runner_cls.__name__}` instance is built from '  # type: ignore # noqa: E501
#             'registry, its implementation can be found in'
#             f'{runner_cls.__module__}',  # type: ignore
#             logger='current',
#             level=logging.DEBUG)
#         return runner

# def build_model_from_cfg(
#     cfg: Union[dict, ConfigDict, Config],
#     registry: Registry,
#     default_args: Optional[Union[dict, 'ConfigDict', 'Config']] = None
# ) -> 'nn.Module':
#     """Build a PyTorch model from config dict(s). Different from
#     ``build_from_cfg``, if cfg is a list, a ``nn.Sequential`` will be built.

#     Args:
#         cfg (dict, list[dict]): The config of modules, which is either a config
#             dict or a list of config dicts. If cfg is a list, the built
#             modules will be wrapped with ``nn.Sequential``.
#         registry (:obj:`Registry`): A registry the module belongs to.
#         default_args (dict, optional): Default arguments to build the module.
#             Defaults to None.

#     Returns:
#         nn.Module: A built nn.Module.
#     """
#     from ..model import Sequential
#     if isinstance(cfg, list):
#         modules = [
#             build_from_cfg(_cfg, registry, default_args) for _cfg in cfg
#         ]
#         return Sequential(*modules)
#     else:
#         return build_from_cfg(cfg, registry, default_args)

# def build_scheduler_from_cfg(
#     cfg: Union[dict, ConfigDict, Config],
#     registry: Registry,
#     default_args: Optional[Union[dict, ConfigDict, Config]] = None
# ) -> '_ParamScheduler':
#     """Builds a ``ParamScheduler`` instance from config.

#     ``ParamScheduler`` supports building instance by its constructor or
#     method ``build_iter_from_epoch``. Therefore, its registry needs a build
#     function to handle both cases.

#     Args:
#         cfg (dict or ConfigDict or Config): Config dictionary. If it contains
#             the key ``convert_to_iter_based``, instance will be built by method
#             ``convert_to_iter_based``, otherwise instance will be built by its
#             constructor.
#         registry (:obj:`Registry`): The ``PARAM_SCHEDULERS`` registry.
#         default_args (dict or ConfigDict or Config, optional): Default
#             initialization arguments. It must contain key ``optimizer``. If
#             ``convert_to_iter_based`` is defined in ``cfg``, it must
#             additionally contain key ``epoch_length``. Defaults to None.

#     Returns:
#         object: The constructed ``ParamScheduler``.
#     """
#     assert isinstance(
#         cfg,
#         (dict, ConfigDict, Config
#          )), f'cfg should be a dict, ConfigDict or Config, but got {type(cfg)}'
#     assert isinstance(
#         registry, Registry), ('registry should be a mmengine.Registry object',
#                               f'but got {type(registry)}')

#     args = cfg.copy()
#     if default_args is not None:
#         for name, value in default_args.items():
#             args.setdefault(name, value)
#     scope = args.pop('_scope_', None)
#     with registry.switch_scope_and_registry(scope) as registry:
#         convert_to_iter = args.pop('convert_to_iter_based', False)
#         if convert_to_iter:
#             scheduler_type = args.pop('type')
#             assert 'epoch_length' in args and args.get('by_epoch', True), (
#                 'Only epoch-based parameter scheduler can be converted to '
#                 'iter-based, and `epoch_length` should be set')
#             if isinstance(scheduler_type, str):
#                 scheduler_cls = registry.get(scheduler_type)
#                 if scheduler_cls is None:
#                     raise KeyError(
#                         f'{scheduler_type} is not in the {registry.name} '
#                         'registry. Please check whether the value of '
#                         f'`{scheduler_type}` is correct or it was registered '
#                         'as expected. More details can be found at https://mmengine.readthedocs.io/en/latest/advanced_tutorials/config.html#import-the-custom-module'  # noqa: E501
#                     )
#             elif inspect.isclass(scheduler_type):
#                 scheduler_cls = scheduler_type
#             else:
#                 raise TypeError('type must be a str or valid type, but got '
#                                 f'{type(scheduler_type)}')
#             return scheduler_cls.build_iter_from_epoch(  # type: ignore
#                 **args)
#         else:
#             args.pop('epoch_length', None)
#             return build_from_cfg(args, registry)



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
import xml.etree.ElementTree as ET
from collections import OrderedDict

import mmcv
import numpy as np
from mmcv import print_log
from mmdet.datasets import CustomDataset
from PIL import Image

from mmrotate.core import eval_rbbox_map, obb2poly_np, poly2obb_np
from mmrotate.datasets.builder import ROTATED_DATASETS


@ROTATED_DATASETS.register_module()
class CustomHRSCDataset(CustomDataset):
    """HRSC dataset for detection.

    Args:
        ann_file (str): Annotation file path.
        pipeline (list[dict]): Processing pipeline.
        img_subdir (str): Subdir where images are stored. Default: JPEGImages.
        ann_subdir (str): Subdir where annotations are. Default: Annotations.
        classwise (bool): Whether to use all classes or only ship.
        version (str, optional): Angle representations. Defaults to 'oc'.
    """

    CLASSES = None
    HRSC_CLASS = ('ship', )
    HRSC_CLASSES = ('ship', 'aircraft carrier', 'warcraft', 'merchant ship',
                    'Nimitz', 'Enterprise', 'Arleigh Burke', 'WhidbeyIsland',
                    'Perry', 'Sanantonio', 'Ticonderoga', 'Kitty Hawk',
                    'Kuznetsov', 'Abukuma', 'Austen', 'Tarawa', 'Blue Ridge',
                    'Container', 'OXo|--)', 'Car carrier([]==[])',
                    'Hovercraft', 'yacht', 'CntShip(_|.--.--|_]=', 'Cruise',
                    'submarine', 'lute', 'Medical', 'Car carrier(======|',
                    'Ford-class', 'Midway-class', 'Invincible-class')
    HRSC_CLASSES_ID = ('01', '02', '03', '04', '05', '06', '07', '08', '09',
                       '10', '11', '12', '13', '14', '15', '16', '17', '18',
                       '19', '20', '22', '24', '25', '26', '27', '28', '29',
                       '30', '31', '32', '33')
    PALETTE = [
        (0, 255, 0),
    ]
    CLASSWISE_PALETTE = [(220, 20, 60), (119, 11, 32),
                         (0, 0, 142), (0, 0, 230), (106, 0, 228), (0, 60, 100),
                         (0, 80, 100), (0, 0, 70), (0, 0, 192), (250, 170, 30),
                         (100, 170, 30), (220, 220, 0), (175, 116, 175),
                         (250, 0, 30), (165, 42, 42), (255, 77, 255),
                         (0, 226, 252), (182, 182, 255), (0, 82, 0),
                         (120, 166, 157), (110, 76, 0), (174, 57, 255),
                         (199, 100, 0), (72, 0, 118), (255, 179, 240),
                         (0, 125, 92), (209, 0, 151), (188, 208, 182),
                         (0, 220, 176), (255, 99, 164), (92, 0, 73)]

    def __init__(self,
                 ann_file,
                 pipeline,
                 img_subdir='JPEGImages',
                 ann_subdir='Annotations',
                 classwise=False,
                 version='oc',
                 # Now iou_thr & use_07_metric are controlled by configs
                 iou_thr=[0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95],
                 use_07_metric=True,
                 **kwargs):
        self.img_subdir = img_subdir
        self.ann_subdir = ann_subdir
        self.classwise = classwise
        self.version = version

        self.iou_thr = iou_thr
        self.use_07_metric = use_07_metric

        if self.classwise:
            HRSCDataset.PALETTE = HRSCDataset.CLASSWISE_PALETTE
            HRSCDataset.CLASSES = self.HRSC_CLASSES
            self.catid2label = {
                ('1' + '0' * 6 + cls_id): i
                for i, cls_id in enumerate(self.HRSC_CLASSES_ID)
            }
        else:
            HRSCDataset.CLASSES = self.HRSC_CLASS
        # self.cat2label = {cat: i for i, cat in enumerate(self.CLASSES)}
        super(HRSCDataset, self).__init__(ann_file, pipeline, **kwargs)

    def load_annotations(self, ann_file):
        """Load annotation from XML style ann_file.

        Args:
            ann_file (str): Path of Imageset file.

        Returns:
            list[dict]: Annotation info from XML file.
        """

        data_infos = []
        img_ids = mmcv.list_from_file(ann_file)
        for img_id in img_ids:
            data_info = {}

            filename = osp.join(self.img_subdir, f'{img_id}.bmp')
            data_info['filename'] = f'{img_id}.bmp'
            xml_path = osp.join(self.ann_subdir, f'{img_id}.xml')   # Delete self.img_prefix,
            tree = ET.parse(xml_path)
            root = tree.getroot()

            width = int(root.find('Img_SizeWidth').text)
            height = int(root.find('Img_SizeHeight').text)

            if width is None or height is None:
                img_path = filename                                 # Delete self.img_prefix,
                img = Image.open(img_path)
                width, height = img.size
            data_info['width'] = width
            data_info['height'] = height
            data_info['ann'] = {}
            gt_bboxes = []
            gt_labels = []
            gt_polygons = []
            gt_headers = []
            gt_bboxes_ignore = []
            gt_labels_ignore = []
            gt_polygons_ignore = []
            gt_headers_ignore = []

            for obj in root.findall('HRSC_Objects/HRSC_Object'):
                if self.classwise:
                    class_id = obj.find('Class_ID').text
                    label = self.catid2label.get(class_id)
                    if label is None:
                        continue
                else:
                    label = 0

                # Add an extra score to use obb2poly_np
                bbox = np.array([[
                    float(obj.find('mbox_cx').text),
                    float(obj.find('mbox_cy').text),
                    float(obj.find('mbox_w').text),
                    float(obj.find('mbox_h').text),
                    float(obj.find('mbox_ang').text), 0
                ]],
                                dtype=np.float32)

                polygon = obb2poly_np(bbox, 'le90')[0, :-1].astype(np.float32)
                if self.version != 'le90':
                    bbox = np.array(
                        poly2obb_np(polygon, self.version), dtype=np.float32)
                else:
                    bbox = bbox[0, :-1]
                head = np.array([
                    int(obj.find('header_x').text),
                    int(obj.find('header_y').text)
                ],
                                dtype=np.int64)

                gt_bboxes.append(bbox)
                gt_labels.append(label)
                gt_polygons.append(polygon)
                gt_headers.append(head)

            if gt_bboxes:
                data_info['ann']['bboxes'] = np.array(
                    gt_bboxes, dtype=np.float32)
                data_info['ann']['labels'] = np.array(
                    gt_labels, dtype=np.int64)
                data_info['ann']['polygons'] = np.array(
                    gt_polygons, dtype=np.float32)
                data_info['ann']['headers'] = np.array(
                    gt_headers, dtype=np.int64)
            else:
                data_info['ann']['bboxes'] = np.zeros((0, 5), dtype=np.float32)
                data_info['ann']['labels'] = np.array([], dtype=np.int64)
                data_info['ann']['polygons'] = np.zeros((0, 8),
                                                        dtype=np.float32)
                data_info['ann']['headers'] = np.zeros((0, 2),
                                                       dtype=np.float32)

            if gt_polygons_ignore:
                data_info['ann']['bboxes_ignore'] = np.array(
                    gt_bboxes_ignore, dtype=np.float32)
                data_info['ann']['labels_ignore'] = np.array(
                    gt_labels_ignore, dtype=np.int64)
                data_info['ann']['polygons_ignore'] = np.array(
                    gt_polygons_ignore, dtype=np.float32)
                data_info['ann']['headers_ignore'] = np.array(
                    gt_headers_ignore, dtype=np.float32)
            else:
                data_info['ann']['bboxes_ignore'] = np.zeros((0, 5),
                                                             dtype=np.float32)
                data_info['ann']['labels_ignore'] = np.array([],
                                                             dtype=np.int64)
                data_info['ann']['polygons_ignore'] = np.zeros(
                    (0, 8), dtype=np.float32)
                data_info['ann']['headers_ignore'] = np.zeros((0, 2),
                                                              dtype=np.float32)

            data_infos.append(data_info)
        return data_infos

    def _filter_imgs(self):
        """Filter images without ground truths."""
        valid_inds = []
        for i, data_info in enumerate(self.data_infos):
            if (not self.filter_empty_gt
                    or data_info['ann']['labels'].size > 0):
                valid_inds.append(i)
        return valid_inds

    def evaluate(
            self,
            results,
            metric='mAP',
            logger=None,
            proposal_nums=(100, 300, 1000),
            iou_thr=[0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95],
            scale_ranges=None,
            use_07_metric=True,
            nproc=4):
        """Evaluate the dataset.

        Args:
            results (list): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated.
            logger (logging.Logger | None | str): Logger used for printing
                related information during evaluation. Default: None.
            proposal_nums (Sequence[int]): Proposal number used for evaluating
                recalls, such as recall@100, recall@1000.
                Default: (100, 300, 1000).
            iou_thr (float | list[float]): IoU threshold. It must be a float
                when evaluating mAP, and can be a list when evaluating recall.
                Default: 0.5.
            scale_ranges (list[tuple] | None): Scale ranges for evaluating mAP.
                Default: None.
            use_07_metric (bool): Whether to use the voc07 metric.
            nproc (int): Processes used for computing TP and FP.
                Default: 4.
        """

        # Now iou_thr & use_07_metric are controlled by configs
        iou_thr = self.iou_thr
        use_07_metric = self.use_07_metric

        if not isinstance(metric, str):
            assert len(metric) == 1
            metric = metric[0]
        allowed_metrics = ['mAP', 'recall']
        if metric not in allowed_metrics:
            raise KeyError(f'metric {metric} is not supported')

        annotations = [self.get_ann_info(i) for i in range(len(self))]
        eval_results = OrderedDict()
        iou_thrs = [iou_thr] if isinstance(iou_thr, float) else iou_thr
        if metric == 'mAP':
            assert isinstance(iou_thrs, list)
            mean_aps = []
            for iou_thr in iou_thrs:
                print_log(f'\n{"-" * 15}iou_thr: {iou_thr}{"-" * 15}')
                mean_ap, _ = eval_rbbox_map(
                    results,
                    annotations,
                    scale_ranges=scale_ranges,
                    iou_thr=iou_thr,
                    use_07_metric=use_07_metric,
                    dataset=self.CLASSES,
                    logger=logger,
                    nproc=nproc)
                mean_aps.append(mean_ap)
                eval_results[f'AP{int(iou_thr * 100):02d}'] = round(mean_ap, 3)
            eval_results['mAP'] = sum(mean_aps) / len(mean_aps)
            eval_results.move_to_end('mAP', last=False)
        elif metric == 'recall':
            raise NotImplementedError

        return eval_results



import os.path as osp
import psutil
import time 

from visionsuite.engines.utils.torch_utils.utils import save_on_master

def on_tester_build_start(tester, *args, **kwargs):
    pass        
     
def on_tester_build_end(tester, *args, **kwargs):
    pass 
    
def on_tester_start(tester, *args, **kwargs):
    pass 

def on_tester_end(tester, *args, **kwargs):
    pass
    
def on_tester_batch_start(tester, *args, **kwargs):
    pass

def on_tester_batch_end(tester, *args, **kwargs):
    pass

def on_tester_step_start(tester, *args, **kwargs): # iteration for a batch
    pass

def on_tester_step_end(tester, *args, **kwargs): # iteration for a batch
    pass

callbacks = {
    "on_tester_build_start": [on_tester_build_start], 
    "on_tester_build_end": [on_tester_build_end], 
    "on_tester_start": [on_tester_start],
    "on_tester_end": [on_tester_end],
    "on_tester_batch_start": [on_tester_batch_start],
    "on_tester_batch_end": [on_tester_batch_end],
    "on_tester_step_start": [on_tester_step_start],
    "on_tester_step_end": [on_tester_step_end],
}




#!/usr/bin/env python3

"""Loading data (images and annotations)."""

import os.path as osp

import numpy as np

import mmcv
from mmseg.datasets.pipelines.loading import LoadAnnotations as MMSEG_LoadAnnotations

from ..builder import PIPELINES


# @PIPELINES.register_module(force=True)
class LoadAnnotations(MMSEG_LoadAnnotations):
    """Extended `LoadAnnotations` class from mmseg.

    Added support for loading instance masks for instance-aware semantic boundaries.
    """

    def __call__(self, results):

        if self.file_client is None:
            self.file_client = mmcv.FileClient(**self.file_client_args)

        if results.get("seg_prefix", None) is not None:
            try:
                filename = osp.join(results["seg_prefix"], results["ann_info"]["seg_map"])
            except:
                filename = osp.join(results["seg_prefix"], results["img_info"]['ann']["seg_map"])
        else:
            filename = results["ann_info"]["seg_map"]
        filename = filename.replace('labelTrainIds', 'labelIds')
        filename = osp.join(results.get("img_prefix"), results.get('img_info')['filename'])
        img_bytes = self.file_client.get(filename)
        gt_semantic_seg = (
            mmcv.imfrombytes(img_bytes, flag="unchanged", backend=self.imdecode_backend)
            .squeeze()
            .astype(np.uint8)
        )
        # modify if custom classes (NOTE: currently not being used)
        if results.get("label_map", None) is not None:
            for old_id, new_id in results["label_map"].items():
                # Add deep copy to solve bug of repeatedly replacing
                # gt_semantic_seg
                gt_semantic_seg_copy = gt_semantic_seg.copy()
                gt_semantic_seg[gt_semantic_seg_copy == old_id] = new_id

        # reduce zero_label
        # It is better to reduce zero label after edges are computed.
        # Unless the it is necessary that the zero label is removed
        if self.reduce_zero_label:
            # avoid using underflow conversion
            gt_semantic_seg[gt_semantic_seg == 0] = 255
            gt_semantic_seg = gt_semantic_seg - 1
            gt_semantic_seg[gt_semantic_seg == 254] = 255

        results["gt_semantic_seg"] = gt_semantic_seg
        results["seg_fields"].append("gt_semantic_seg")

        # load instance map
        inst_sensitive = results.get("inst_sensitive", False)
        if inst_sensitive:
            # load instance segmentation image
            inst_prefix = results.get("inst_prefix", None)
            if inst_prefix is not None:
                filename = osp.join(inst_prefix, results["ann_info"]["inst_map"])
            else:
                filename = results["ann_info"]["inst_map"]
            img_bytes = self.file_client.get(filename)
            gt_inst_seg = (
                mmcv.imfrombytes(
                    img_bytes, flag="unchanged", backend=self.imdecode_backend
                )
                .squeeze()
                .astype(np.int32)  # NOTE: needs to be int32
            )
            results["gt_inst_seg"] = gt_inst_seg
            results["seg_fields"].append("gt_inst_seg")

        return results



import os 
import warnings
import datetime
import os.path as osp
from glob import glob 
from pathlib import Path
import re

def increment_path(path, exist_ok=False, sep="", mkdir=False):
    # Increment file or directory path, i.e. runs/exp --> runs/exp{sep}2, runs/exp{sep}3, ... etc.
    path = Path(path)  # os-agnostic
    if path.exists() and not exist_ok:
        path, suffix = (
            (path.with_suffix(""), path.suffix) if path.is_file() else (path, "")
        )
        dirs = glob(f"{path}{sep}*")  # similar paths
        matches = [re.search(rf"%s{sep}(\d+)" % path.stem, d) for d in dirs]
        i = [int(m.groups()[0]) for m in matches if m]  # indices
        n = max(i) + 1 if i else 2  # increment number
        path = Path(f"{path}{sep}{n}{suffix}")  # increment path
    if mkdir:
        path.mkdir(parents=True, exist_ok=True)  # make directory
    return path



def print_class_name_on_instantiation(cls):
    original_init = cls.__init__

    def new_init(self, *args, **kwargs):
        print(f"Instantiating class: {cls.__name__}")
        original_init(self, *args, **kwargs)

    cls.__init__ = new_init
    return cls

def mkdir(path, make_dirs=False):
    if not osp.exists(path):
        if make_dirs:
            os.makedirs(path)
        else:
            os.mkdir(path)
            
        print(f"Created directory: {path}")
    else:
        warnings.warn(f"There is already {path}")

def get_cache_path(filepath):
    import hashlib

    h = hashlib.sha1(filepath.encode()).hexdigest()
    cache_path = os.path.join("~", ".torch", "vision", "datasets", "imagefolder", h[:10] + ".pt")
    cache_path = os.path.expanduser(cache_path)
    return cache_path

def create_output_dir(output_dir, mode, make_dirs=False):
    now = datetime.datetime.now()
    year = now.year 
    month = now.month
    day = now.day
    hour = now.hour
    minute = now.minute
    second = now.second
    
    output_dir = osp.join(output_dir, f'{year}_{month}_{day}_{hour}_{minute}_{second}', mode)
    mkdir(output_dir, make_dirs=make_dirs)
    
    return output_dir
    
def yaml2namespace(args_file):
    from types import SimpleNamespace
    import yaml 
    
    with open(args_file, 'r') as yf:
        cfgs = yaml.safe_load(yf)
        
    args = SimpleNamespace(**cfgs)
    
    return args
    
def yaml2dict(args_file):
    import yaml 
    
    with open(args_file, 'r') as yf:
        cfgs = yaml.safe_load(yf)
        
    return cfgs

def assert_key_dict(dictionary, key):
    assert key in dictionary, ValueError(f"There is no key({key})")
    
def get_params_from_obj(obj):
    import inspect

    signature = inspect.signature(obj)

    # Extract the parameters
    parameters = {param.name: param.default for param in signature.parameters.values()}
    
    return parameters

def update_dict(A, B):
    for key, value in B.items():
        # assert key in A, ValueError(f"There is no such key({key})")
        if isinstance(value, dict) and key in A and isinstance(A[key], dict):
            update_dict(A[key], value)
        else:
            A[key] = value


import glob 
import os
import os.path as osp
import torch
import imgviz
import json
import cv2
import numpy as np
import pandas as pd
from tqdm import tqdm

from mlearning.utils.vis.vis_hbb import vis_hbb
from mlearning.utils.functionals import letterbox
from athena.src.tasks.detection.frameworks.pytorch.models.yolov7.models.experimental import attempt_load as attempt_load7
from athena.src.tasks.detection.frameworks.pytorch.models.yolov5.utils.general import check_img_size
from athena.src.tasks.detection.frameworks.pytorch.models.yolov5.utils.general import (non_max_suppression,
                                                                                       scale_coords)

compare_mask = True
imgsz = 2048
device = 'cuda'

# weights = '/DeepLearning/_projects/sungjin_body/tr_0.1/train/weights/best.pt'

# # input_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/w_json/기타'
# # json_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/w_json/기타'
# # output_dir = f'/DeepLearning/_projects/sungjin_body/tests/athena_tr_0.1/winter/w_json/기타'

# input_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/wo_json/기타'
# json_dir = None
# output_dir = f'/DeepLearning/_projects/sungjin_body/tests/athena_tr_0.1/winter/wo_json/기타'


weights = '/DeepLearning/_projects/sungjin_body/tr_0.5/2024_07_29_04_58_53/train/weights/best.pt'

# input_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/w_json/기타'
# json_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/w_json/기타'
# output_dir = f'/DeepLearning/_projects/sungjin_body/tests/athena_tr_0.5/winter/w_json/기타'

input_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/wo_json/학습'
json_dir = None
output_dir = f'/DeepLearning/_projects/sungjin_body/tests/athena_tr_0.5/winter/wo_json/학습'

compare_gt = True if json_dir is not None else False
model = attempt_load7(weights, map_location=device)  # load FP32 model
imgsz = check_img_size(imgsz, s=model.stride.max())  # check image size
if isinstance(imgsz, int):
    imgsz = (imgsz, imgsz)
model(torch.zeros(1, 3, imgsz[0], imgsz[1]).to(device).type_as(next(model.parameters())))  # run once
idx2class = {ii: vv for ii, vv in enumerate(model.classes)}

iou_threshold = 0.5
max_dets = 50
nms_conf_threshold = 0.25
nms_iou_threshold = 0.1
line_width = 5
font_scale = 2
classes = model.classes

if not osp.exists(output_dir):
    os.makedirs(output_dir)

img_files = glob.glob(osp.join(input_dir, '*.bmp'))

results = {}
compare = {}
for img_file in tqdm(img_files):
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    img = cv2.imread(img_file).astype(np.float32)
    img_h, img_w, img_c = img.shape
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)/255
    img, ratio, pad_values = letterbox(img, (imgsz[0], imgsz[1]))
    
    torch_img = torch.from_numpy(np.transpose(img[np.newaxis, :], (0, 3, 1, 2))).to('cuda')
    with torch.no_grad():
        preds = model(torch_img)
    pred = preds[0].detach().cpu()
    dets = non_max_suppression(pred, nms_conf_threshold, nms_iou_threshold, None, False)
    
    idx2xyxys = {}
    for det in dets:
        if len(det):
            det[:, :4] = scale_coords(img.shape[:2], det[:, :4], (img_h, img_w)).round()

            for *xyxy, conf, cls in reversed(det):
                cls = cls.detach().cpu().item()
                conf = float(conf.detach().cpu().item())
                if cls not in idx2xyxys.keys():
                    idx2xyxys[cls] = []
                    idx2xyxys[cls] = {'bbox': [], 'confidence': 0}
                    
                idx2xyxys[cls]['bbox'].append([[int(np.round(xyxy[0].detach().cpu().item())), int(np.round(xyxy[1].detach().cpu().item()))], 
                                       [int(np.round(xyxy[2].detach().cpu().item())), int(np.round(xyxy[3].detach().cpu().item()))]])
                idx2xyxys[cls]['confidence'] = conf
  
    color_map = imgviz.label_colormap()[1:len(idx2class) + 1 + 1]
    results.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
    
    if compare_gt:
        _compare = vis_hbb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
                           compare_gt=compare_gt, iou_threshold=iou_threshold, line_width=line_width, font_scale=font_scale)
        _compare.update({"img_file": img_file})
        compare.update({filename: _compare})
    else:
        vis_hbb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
                compare_gt=compare_gt, iou_threshold=iou_threshold, line_width=line_width, font_scale=font_scale)
            
with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
    json.dump(results, json_file, ensure_ascii=False, indent=4)

if compare_gt:
    with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
        json.dump(compare, json_file, ensure_ascii=False, indent=4)
    
    df_compare = pd.DataFrame(compare)
    df_compare_pixel = df_compare.loc['diff_iou'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))


import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from functools import partial
import torch.utils.checkpoint as checkpoint
from timm.models.vision_transformer import DropPath, Mlp, trunc_normal_
from timm.models.layers import to_2tuple


class Attention(nn.Module):
    def __init__(self, input_size, dim, num_heads, qkv_bias=True, qk_scale=None, 
                 attn_drop=0., proj_drop=0., rpe=True):
        super().__init__()
        self.input_size = input_size
        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        # define a parameter table of relative position bias
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * input_size - 1) * (2 * input_size - 1), num_heads)
        ) if rpe else None
        # if rpe:
        #     trunc_normal_(self.relative_position_bias_table, std=.02)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, rpe_index=None, mask=None):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy (cannot use tensor as tuple)

        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))

        if rpe_index is not None:
            S = int(math.sqrt(rpe_index.size(-1)))
            relative_position_bias = self.relative_position_bias_table[rpe_index].view(-1, S, S, self.num_heads)
            relative_position_bias = relative_position_bias.permute(0, 3, 1, 2).contiguous()
            assert N == S
            attn = attn + relative_position_bias
        if mask is not None:
            mask = mask.bool()
            attn = attn.masked_fill(~mask[:, None, None, :], float("-inf"))
        attn = attn.float().clamp(min=torch.finfo(torch.float32).min, max=torch.finfo(torch.float32).max)
        attn = self.softmax(attn)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class BlockWithRPE(nn.Module):
    def __init__(self, input_size, dim, num_heads=0., mlp_ratio=4., qkv_bias=True, qk_scale=None, 
                 drop=0., attn_drop=0., drop_path=0., rpe=True,
                 act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio

        with_attn = num_heads > 0.

        self.norm1 = norm_layer(dim) if with_attn else None
        self.attn = Attention(
            input_size, dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, 
            attn_drop=attn_drop, proj_drop=drop, rpe=rpe,
        ) if with_attn else None

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x, rpe_index=None, mask=None):
        if self.attn is not None:
            x = x + self.drop_path(self.attn(self.norm1(x), rpe_index, mask))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x
 

class PatchEmbed(nn.Module):
    def __init__(self, img_size=224, patch_size=16, inner_patches=4, in_chans=3, embed_dim=96, norm_layer=None):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        patches_resolution = [img_size[0] // patch_size[0], img_size[1] // patch_size[1]]
        self.img_size = img_size
        self.patch_size = patch_size
        self.inner_patches = inner_patches
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        conv_size = [size // inner_patches for size in patch_size]
        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=conv_size, stride=conv_size)
        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    def forward(self, x):
        B, C, H, W = x.shape
        patches_resolution = (H // self.patch_size[0], W // self.patch_size[1])
        num_patches = patches_resolution[0] * patches_resolution[1]
        x = self.proj(x).view(
            B, -1, 
            patches_resolution[0], self.inner_patches, 
            patches_resolution[1], self.inner_patches, 
        ).permute(0, 2, 4, 3, 5, 1).reshape(B, num_patches, self.inner_patches, self.inner_patches, -1)
        if self.norm is not None:
            x = self.norm(x)
        return x


class PatchMerge(nn.Module):
    def __init__(self, dim, norm_layer):
        super().__init__()
        self.norm = norm_layer(dim * 4)
        self.reduction = nn.Linear(dim * 4, dim * 2, bias=False)
    
    def forward(self, x):
        x0 = x[..., 0::2, 0::2, :] 
        x1 = x[..., 1::2, 0::2, :] 
        x2 = x[..., 0::2, 1::2, :] 
        x3 = x[..., 1::2, 1::2, :]

        x = torch.cat([x0, x1, x2, x3], dim=-1)
        x = self.norm(x)
        x = self.reduction(x)
        return x


class HiViT(nn.Module):
    def __init__(self, img_size=224, patch_size=16, in_chans=3, num_classes=1000,
                 embed_dim=512, depths=[4, 4, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
                 qkv_bias=True, qk_scale=None, drop_rate=0., attn_drop_rate=0., drop_path_rate=0.0,
                 norm_layer=nn.LayerNorm, ape=True, rpe=True, patch_norm=True, use_checkpoint=False, 
                 **kwargs):
        super().__init__()
        self.num_classes = num_classes
        self.num_layers = len(depths)
        self.ape = ape
        self.rpe = rpe
        self.patch_norm = patch_norm
        self.num_features = embed_dim
        self.mlp_ratio = mlp_ratio
        self.use_checkpoint = use_checkpoint
        self.num_main_blocks = depths[-1]

        embed_dim = embed_dim // 2 ** (self.num_layers - 1)
        # split image into non-overlapping patches
        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)
        num_patches = self.patch_embed.num_patches
        Hp, Wp = self.patch_embed.patches_resolution
        assert Hp == Wp

        # absolute position embedding
        if ape:
            self.absolute_pos_embed = nn.Parameter(
                torch.zeros(1, num_patches, self.num_features)
            )
            trunc_normal_(self.absolute_pos_embed, std=.02)
        if rpe:
            coords_h = torch.arange(Hp)
            coords_w = torch.arange(Wp)
            coords = torch.stack(torch.meshgrid([coords_h, coords_w])) 
            coords_flatten = torch.flatten(coords, 1) 
            relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :] 
            relative_coords = relative_coords.permute(1, 2, 0).contiguous() 
            relative_coords[:, :, 0] += Hp - 1 
            relative_coords[:, :, 1] += Wp - 1
            relative_coords[:, :, 0] *= 2 * Wp - 1
            relative_position_index = relative_coords.sum(-1)
            self.register_buffer("relative_position_index", relative_position_index)

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = iter(x.item() for x in torch.linspace(0, drop_path_rate, sum(depths) + sum(depths[:-1])))  # stochastic depth decay rule

        # build blocks
        self.blocks = nn.ModuleList()
        for stage_depth in depths:
            is_main_stage = embed_dim == self.num_features
            nhead = num_heads if is_main_stage else 0
            ratio = mlp_ratio if is_main_stage else stem_mlp_ratio
            # every block not in main stage include two mlp blocks
            stage_depth = stage_depth if is_main_stage else stage_depth * 2
            for i in range(stage_depth):
                self.blocks.append(
                    BlockWithRPE(
                        Hp, embed_dim, nhead, ratio, qkv_bias, qk_scale, 
                        drop=drop_rate, attn_drop=attn_drop_rate, drop_path=next(dpr), 
                        rpe=rpe, norm_layer=norm_layer,
                    )
                )
            if not is_main_stage:
                self.blocks.append(
                    PatchMerge(embed_dim, norm_layer)
                )
                embed_dim *= 2

        self.fc_norm = norm_layer(self.num_features)
        self.head = nn.Linear(self.num_features, num_classes) if num_classes > 0 else nn.Identity()

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'absolute_pos_embed'}

    @torch.jit.ignore
    def no_weight_decay_keywords(self):
        return {'relative_position_bias_table'}

    def forward_features(self, x, ids_keep=None, mask=None):
        B = x.shape[0]
        x = self.patch_embed(x)
        if ids_keep is not None:
            x = torch.gather(
                x, dim=1, index=ids_keep[:, :, None, None, None].expand(-1, -1, *x.shape[2:])
            )
        
        for blk in self.blocks[:-self.num_main_blocks]:
            x = checkpoint.checkpoint(blk, x) if self.use_checkpoint else blk(x)
        x = x[..., 0, 0, :]
        if self.ape:
            pos_embed = self.absolute_pos_embed
            if ids_keep is not None:
                pos_embed = torch.gather(
                    pos_embed.expand(B, -1, -1), 
                    dim=1, 
                    index=ids_keep[:, :, None].expand(-1, -1, pos_embed.shape[2]),
                )
            x += pos_embed
        x = self.pos_drop(x)

        rpe_index = None
        if self.rpe:
            if ids_keep is not None:
                B, L = ids_keep.shape
                rpe_index = self.relative_position_index
                rpe_index = torch.gather(
                    rpe_index[ids_keep, :], dim=-1, index=ids_keep[:, None, :].expand(-1, L, -1)
                ).reshape(B, -1)
            else:
                rpe_index = self.relative_position_index.view(-1)
        
        for blk in self.blocks[-self.num_main_blocks:]:
            x = checkpoint.checkpoint(blk, x, rpe_index, mask) if self.use_checkpoint else blk(x, rpe_index, mask)
        return x

    def forward(self, x):
        x = self.forward_features(x)
        x = x.mean(dim=1)
        x = self.fc_norm(x)
        x = self.head(x)
        return x


def hivit_base(**kwargs):
    model = HiViT(
        embed_dim=512, depths=[4, 4, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
        rpe=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model



from visionsuite.engines.classification.utils.registry import VALIDATORS

def build_validator(**val_config):
    validator = VALIDATORS.get(val_config['type'], case_sensitive=True)
    
    return validator


import re 
from pathlib import Path 

from setuptools import setup, find_packages

FILE = Path(__file__).resolve()
PARENT = FILE.parent 
README = (PARENT / 'README.md').read_text(encoding='utf-8')

def get_version():
    file = PARENT / 'visionsuite/__init__.py'
    
    return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', file.read_text(encoding='utf-8'), re.M)[1]

def parse_requirements(file_path: Path):
    requirements = []
    for line in Path(file_path).read_text().splitlines():
        line = line.strip()
        
        if line and not line.startswith('#'):
            requirements.append(line.split("#")[0].strip())
    
    return requirements 

try:
    setup(
        name='visionsuite',
        version='1.0.16',
        python_requires='>=3.9',
        description=('Template for Python Athena'),
        long_description=README,
        long_description_content_type='text/markdown',
        # packages=['athena'] + [str(x) for x in Path('athena').rglob('*/') if x.is_dir() and '__' not in str(x)],
        packages=find_packages(exclude=[]),
        package_data={
            '': ['*.yaml', '*.json'], },
        include_package_data=True,
        install_requires=parse_requirements(PARENT / 'requirements.txt') + ["nvidia-ml-py3"],
    )
except:
    setup(
        name='visionsuite',
        version=get_version(),
        python_requires='>=3.9',
        description=('Template for Python Athena'),
        long_description=README,
        long_description_content_type='text/markdown',
        # packages=['athena'] + [str(x) for x in Path('athena').rglob('*/') if x.is_dir() and '__' not in str(x)],
        packages=find_packages(exclude=[]),
        package_data={
            '': ['*.yaml', '*.json'], },
        include_package_data=True,
        # install_requires=["nvidia-ml-py3"],
        install_requires=parse_requirements(PARENT / 'requirements.txt') + ["nvidia-ml-py3"],
    )


import torch
from .dist import is_main_process

def load_ckpt(ckpt, map_location='cpu', weights_only=False):
    try:
        return torch.load(ckpt, map_location=map_location, weights_only=weights_only)
    except Exception as error:
        raise Exception(f"There has been error when loading checkpoint({ckpt}): {error}")


def get_device(device):
    return torch.device(device)

def set_torch_deterministic(use_deterministic_algorithms):
    if use_deterministic_algorithms:
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)
    else:
        torch.backends.cudnn.benchmark = True

def set_test_setting():
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True

def cat_list(images, fill_value=0):
    max_size = tuple(max(s) for s in zip(*[img.shape for img in images]))
    batch_shape = (len(images),) + max_size
    batched_imgs = images[0].new(*batch_shape).fill_(fill_value)
    for img, pad_img in zip(images, batched_imgs):
        pad_img[..., : img.shape[-2], : img.shape[-1]].copy_(img)
    return batched_imgs


def collate_fn(batch):
    if len(batch) == 3:
        images, targets, filenames = list(zip(*batch))
        
        batched_imgs = cat_list(images, fill_value=0)
        batched_targets = cat_list(targets, fill_value=255)
        return batched_imgs, batched_targets, filenames
    elif len(batch) == 2:
        images, targets = list(zip(*batch))
        
        batched_imgs = cat_list(images, fill_value=0)
        batched_targets = cat_list(targets, fill_value=255)
        return batched_imgs, batched_targets
    else:
        raise RuntimeError(f"You need to add output arguments at dataset. There are {len(*batch)} outputs")
        

def save_on_master(*args, **kwargs):
    if is_main_process():
        torch.save(*args, **kwargs)

def parse_device_ids(device_ids):
    if isinstance(device_ids, str):
        return [int(id.strip()) for id in device_ids.split(',')]
    elif isinstance(device_ids, (int, float)):
        return [int(device_ids)]
    elif isinstance(device_ids, list):
        return list(map(int, device_ids))
    else:
        NotImplementedError(f"Not consider this case: {type(device_ids)} for {device_ids}")

        
import copy
import hashlib
import os
from collections import OrderedDict
from typing import List, Optional, Tuple

import torch

def average_checkpoints(inputs):
    """Loads checkpoints from inputs and returns a model with averaged weights. Original implementation taken from:
    https://github.com/pytorch/fairseq/blob/a48f235636557b8d3bc4922a6fa90f3a0fa57955/scripts/average_checkpoints.py#L16

    Args:
      inputs (List[str]): An iterable of string paths of checkpoints to load from.
    Returns:
      A dict of string keys mapping to various values. The 'model' key
      from the returned dict should correspond to an OrderedDict mapping
      string parameter names to torch Tensors.
    """
    params_dict = OrderedDict()
    params_keys = None
    new_state = None
    num_models = len(inputs)
    for fpath in inputs:
        with open(fpath, "rb") as f:
            state = torch.load(
                f, map_location=(lambda s, _: torch.serialization.default_restore_location(s, "cpu")), weights_only=True
            )
        # Copies over the settings from the first checkpoint
        if new_state is None:
            new_state = state
        model_params = state["model"]
        model_params_keys = list(model_params.keys())
        if params_keys is None:
            params_keys = model_params_keys
        elif params_keys != model_params_keys:
            raise KeyError(
                f"For checkpoint {f}, expected list of params: {params_keys}, but found: {model_params_keys}"
            )
        for k in params_keys:
            p = model_params[k]
            if isinstance(p, torch.HalfTensor):
                p = p.float()
            if k not in params_dict:
                params_dict[k] = p.clone()
                # NOTE: clone() is needed in case of p is a shared parameter
            else:
                params_dict[k] += p
    averaged_params = OrderedDict()
    for k, v in params_dict.items():
        averaged_params[k] = v
        if averaged_params[k].is_floating_point():
            averaged_params[k].div_(num_models)
        else:
            averaged_params[k] //= num_models
    new_state["model"] = averaged_params
    return new_state


def store_model_weights(model, checkpoint_path, checkpoint_key="model", strict=True):
    """
    This method can be used to prepare weights files for new models. It receives as
    input a model architecture and a checkpoint from the training script and produces
    a file with the weights ready for release.

    Examples:
        from torchvision import models as M

        # Classification
        model = M.mobilenet_v3_large(weights=None)
        print(store_model_weights(model, './class.pth'))

        # Quantized Classification
        model = M.quantization.mobilenet_v3_large(weights=None, quantize=False)
        model.fuse_model(is_qat=True)
        model.qconfig = torch.ao.quantization.get_default_qat_qconfig('qnnpack')
        _ = torch.ao.quantization.prepare_qat(model, inplace=True)
        print(store_model_weights(model, './qat.pth'))

        # Object Detection
        model = M.detection.fasterrcnn_mobilenet_v3_large_fpn(weights=None, weights_backbone=None)
        print(store_model_weights(model, './obj.pth'))

        # Segmentation
        model = M.segmentation.deeplabv3_mobilenet_v3_large(weights=None, weights_backbone=None, aux_loss=True)
        print(store_model_weights(model, './segm.pth', strict=False))

    Args:
        model (pytorch.nn.Module): The model on which the weights will be loaded for validation purposes.
        checkpoint_path (str): The path of the checkpoint we will load.
        checkpoint_key (str, optional): The key of the checkpoint where the model weights are stored.
            Default: "model".
        strict (bool): whether to strictly enforce that the keys
            in :attr:`state_dict` match the keys returned by this module's
            :meth:`~torch.nn.Module.state_dict` function. Default: ``True``

    Returns:
        output_path (str): The location where the weights are saved.
    """
    # Store the new model next to the checkpoint_path
    checkpoint_path = os.path.abspath(checkpoint_path)
    output_dir = os.path.dirname(checkpoint_path)

    # Deep copy to avoid side effects on the model object.
    model = copy.deepcopy(model)
    checkpoint = torch.load(checkpoint_path, map_location="cpu", weights_only=True)

    # Load the weights to the model to validate that everything works
    # and remove unnecessary weights (such as auxiliaries, etc.)
    if checkpoint_key == "model_ema":
        del checkpoint[checkpoint_key]["n_averaged"]
        torch.nn.modules.utils.consume_prefix_in_state_dict_if_present(checkpoint[checkpoint_key], "module.")
    model.load_state_dict(checkpoint[checkpoint_key], strict=strict)

    tmp_path = os.path.join(output_dir, str(model.__hash__()))
    torch.save(model.state_dict(), tmp_path)

    sha256_hash = hashlib.sha256()
    with open(tmp_path, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
        hh = sha256_hash.hexdigest()

    output_path = os.path.join(output_dir, "weights-" + str(hh[:8]) + ".pth")
    os.replace(tmp_path, output_path)

    return output_path


def set_weight_decay(
    model: torch.nn.Module,
    weight_decay: float,
    norm_weight_decay: Optional[float] = None,
    norm_classes: Optional[List[type]] = None,
    custom_keys_weight_decay: Optional[List[Tuple[str, float]]] = None,
):
    if not norm_classes:
        norm_classes = [
            torch.nn.modules.batchnorm._BatchNorm,
            torch.nn.LayerNorm,
            torch.nn.GroupNorm,
            torch.nn.modules.instancenorm._InstanceNorm,
            torch.nn.LocalResponseNorm,
        ]
    norm_classes = tuple(norm_classes)

    params = {
        "other": [],
        "norm": [],
    }
    params_weight_decay = {
        "other": weight_decay,
        "norm": norm_weight_decay,
    }
    custom_keys = []
    if custom_keys_weight_decay is not None:
        for key, weight_decay in custom_keys_weight_decay:
            params[key] = []
            params_weight_decay[key] = weight_decay
            custom_keys.append(key)

    def _add_params(module, prefix=""):
        for name, p in module.named_parameters(recurse=False):
            if not p.requires_grad:
                continue
            is_custom_key = False
            for key in custom_keys:
                target_name = f"{prefix}.{name}" if prefix != "" and "." in key else name
                if key == target_name:
                    params[key].append(p)
                    is_custom_key = True
                    break
            if not is_custom_key:
                if norm_weight_decay is not None and isinstance(module, norm_classes):
                    params["norm"].append(p)
                else:
                    params["other"].append(p)

        for child_name, child_module in module.named_children():
            child_prefix = f"{prefix}.{child_name}" if prefix != "" else child_name
            _add_params(child_module, prefix=child_prefix)

    _add_params(model)

    param_groups = []
    for key in params:
        if len(params[key]) > 0:
            param_groups.append({"params": params[key], "weight_decay": params_weight_decay[key]})
    return param_groups



from visionsuite.engines.classification.utils.registry import MODELS

def build_model(**config):
    model = MODELS.get(config['type'], 
                       case_sensitive=config['case_sensitive'])()
    
    return model




from __future__ import absolute_import

from keras_unet_collection.activations import GELU, Snake
from tensorflow import expand_dims
from tensorflow.compat.v1 import image
from tensorflow.keras.layers import MaxPooling2D, AveragePooling2D, UpSampling2D, Conv2DTranspose, GlobalAveragePooling2D
from tensorflow.keras.layers import Conv2D, DepthwiseConv2D, Lambda
from tensorflow.keras.layers import BatchNormalization, Activation, concatenate, multiply, add
from tensorflow.keras.layers import ReLU, LeakyReLU, PReLU, ELU, Softmax

def decode_layer(X, channel, pool_size, unpool, kernel_size=3, 
                 activation='ReLU', batch_norm=False, name='decode'):
    '''
    An overall decode layer, based on either upsampling or trans conv.
    
    decode_layer(X, channel, pool_size, unpool, kernel_size=3,
                 activation='ReLU', batch_norm=False, name='decode')
    
    Input
    ----------
        X: input tensor.
        pool_size: the decoding factor.
        channel: (for trans conv only) number of convolution filters.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.           
        kernel_size: size of convolution kernels. 
                     If kernel_size='auto', then it equals to the `pool_size`.
        activation: one of the `tensorflow.keras.layers` interface, e.g., ReLU.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
    
    * The defaut: `kernel_size=3`, is suitable for `pool_size=2`.
    
    '''
    # parsers
    if unpool is False:
        # trans conv configurations
        bias_flag = not batch_norm
    
    elif unpool == 'nearest':
        # upsample2d configurations
        unpool = True
        interp = 'nearest'
    
    elif (unpool is True) or (unpool == 'bilinear'):
        # upsample2d configurations
        unpool = True
        interp = 'bilinear'
    
    else:
        raise ValueError('Invalid unpool keyword')
        
    if unpool:
        X = UpSampling2D(size=(pool_size, pool_size), interpolation=interp, name='{}_unpool'.format(name))(X)
    else:
        if kernel_size == 'auto':
            kernel_size = pool_size
            
        X = Conv2DTranspose(channel, kernel_size, strides=(pool_size, pool_size), 
                            padding='same', name='{}_trans_conv'.format(name))(X)
        
        # batch normalization
        if batch_norm:
            X = BatchNormalization(axis=3, name='{}_bn'.format(name))(X)
            
        # activation
        if activation is not None:
            activation_func = eval(activation)
            X = activation_func(name='{}_activation'.format(name))(X)
        
    return X

def encode_layer(X, channel, pool_size, pool, kernel_size='auto', 
                 activation='ReLU', batch_norm=False, name='encode'):
    '''
    An overall encode layer, based on one of the:
    (1) max-pooling, (2) average-pooling, (3) strided conv2d.
    
    encode_layer(X, channel, pool_size, pool, kernel_size='auto', 
                 activation='ReLU', batch_norm=False, name='encode')
    
    Input
    ----------
        X: input tensor.
        pool_size: the encoding factor.
        channel: (for strided conv only) number of convolution filters.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        kernel_size: size of convolution kernels. 
                     If kernel_size='auto', then it equals to the `pool_size`.
        activation: one of the `tensorflow.keras.layers` interface, e.g., ReLU.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
        
    '''
    # parsers
    if (pool in [False, True, 'max', 'ave']) is not True:
        raise ValueError('Invalid pool keyword')
        
    # maxpooling2d as default
    if pool is True:
        pool = 'max'
        
    elif pool is False:
        # stride conv configurations
        bias_flag = not batch_norm
    
    if pool == 'max':
        X = MaxPooling2D(pool_size=(pool_size, pool_size), name='{}_maxpool'.format(name))(X)
        
    elif pool == 'ave':
        X = AveragePooling2D(pool_size=(pool_size, pool_size), name='{}_avepool'.format(name))(X)
        
    else:
        if kernel_size == 'auto':
            kernel_size = pool_size
        
        # linear convolution with strides
        X = Conv2D(channel, kernel_size, strides=(pool_size, pool_size), 
                   padding='valid', use_bias=bias_flag, name='{}_stride_conv'.format(name))(X)
        
        # batch normalization
        if batch_norm:
            X = BatchNormalization(axis=3, name='{}_bn'.format(name))(X)
            
        # activation
        if activation is not None:
            activation_func = eval(activation)
            X = activation_func(name='{}_activation'.format(name))(X)
            
    return X

def attention_gate(X, g, channel,  
                   activation='ReLU', 
                   attention='add', name='att'):
    '''
    Self-attention gate modified from Oktay et al. 2018.
    
    attention_gate(X, g, channel,  activation='ReLU', attention='add', name='att')
    
    Input
    ----------
        X: input tensor, i.e., key and value.
        g: gated tensor, i.e., query.
        channel: number of intermediate channel.
                 Oktay et al. (2018) did not specify (denoted as F_int).
                 intermediate channel is expected to be smaller than the input channel.
        activation: a nonlinear attnetion activation.
                    The `sigma_1` in Oktay et al. 2018. Default is 'ReLU'.
        attention: 'add' for additive attention; 'multiply' for multiplicative attention.
                   Oktay et al. 2018 applied additive attention.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X_att: output tensor.
    
    '''
    activation_func = eval(activation)
    attention_func = eval(attention)
    
    # mapping the input tensor to the intermediate channel
    theta_att = Conv2D(channel, 1, use_bias=True, name='{}_theta_x'.format(name))(X)
    
    # mapping the gate tensor
    phi_g = Conv2D(channel, 1, use_bias=True, name='{}_phi_g'.format(name))(g)
    
    # ----- attention learning ----- #
    query = attention_func([theta_att, phi_g], name='{}_add'.format(name))
    
    # nonlinear activation
    f = activation_func(name='{}_activation'.format(name))(query)
    
    # linear transformation
    psi_f = Conv2D(1, 1, use_bias=True, name='{}_psi_f'.format(name))(f)
    # ------------------------------ #
    
    # sigmoid activation as attention coefficients
    coef_att = Activation('sigmoid', name='{}_sigmoid'.format(name))(psi_f)
    
    # multiplicative attention masking
    X_att = multiply([X, coef_att], name='{}_masking'.format(name))
    
    return X_att

def CONV_stack(X, channel, kernel_size=3, stack_num=2, 
               dilation_rate=1, activation='ReLU', 
               batch_norm=False, name='conv_stack'):
    '''
    Stacked convolutional layers:
    (Convolutional layer --> batch normalization --> Activation)*stack_num
    
    CONV_stack(X, channel, kernel_size=3, stack_num=2, dilation_rate=1, activation='ReLU', 
               batch_norm=False, name='conv_stack')
    
    
    Input
    ----------
        X: input tensor.
        channel: number of convolution filters.
        kernel_size: size of 2-d convolution kernels.
        stack_num: number of stacked Conv2D-BN-Activation layers.
        dilation_rate: optional dilated convolution kernel.
        activation: one of the `tensorflow.keras.layers` interface, e.g., ReLU.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor
        
    '''
    
    bias_flag = not batch_norm
    
    # stacking Convolutional layers
    for i in range(stack_num):
        
        activation_func = eval(activation)
        
        # linear convolution
        X = Conv2D(channel, kernel_size, padding='same', use_bias=bias_flag, 
                   dilation_rate=dilation_rate, name='{}_{}'.format(name, i))(X)
        
        # batch normalization
        if batch_norm:
            X = BatchNormalization(axis=3, name='{}_{}_bn'.format(name, i))(X)
        
        # activation
        activation_func = eval(activation)
        X = activation_func(name='{}_{}_activation'.format(name, i))(X)
        
    return X

def Res_CONV_stack(X, X_skip, channel, res_num, activation='ReLU', batch_norm=False, name='res_conv'):
    '''
    Stacked convolutional layers with residual path.
     
    Res_CONV_stack(X, X_skip, channel, res_num, activation='ReLU', batch_norm=False, name='res_conv')
     
    Input
    ----------
        X: input tensor.
        X_skip: the tensor that does go into the residual path 
                can be a copy of X (e.g., the identity block of ResNet).
        channel: number of convolution filters.
        res_num: number of convolutional layers within the residual path.
        activation: one of the `tensorflow.keras.layers` interface, e.g., 'ReLU'.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
        
    '''  
    X = CONV_stack(X, channel, kernel_size=3, stack_num=res_num, dilation_rate=1, 
                   activation=activation, batch_norm=batch_norm, name=name)

    X = add([X_skip, X], name='{}_add'.format(name))
    
    activation_func = eval(activation)
    X = activation_func(name='{}_add_activation'.format(name))(X)
    
    return X

def Sep_CONV_stack(X, channel, kernel_size=3, stack_num=1, dilation_rate=1, activation='ReLU', batch_norm=False, name='sep_conv'):
    '''
    Depthwise separable convolution with (optional) dilated convolution kernel and batch normalization.
    
    Sep_CONV_stack(X, channel, kernel_size=3, stack_num=1, dilation_rate=1, activation='ReLU', batch_norm=False, name='sep_conv')
    
    Input
    ----------
        X: input tensor.
        channel: number of convolution filters.
        kernel_size: size of 2-d convolution kernels.
        stack_num: number of stacked depthwise-pointwise layers.
        dilation_rate: optional dilated convolution kernel.
        activation: one of the `tensorflow.keras.layers` interface, e.g., 'ReLU'.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
    
    '''
    
    activation_func = eval(activation)
    bias_flag = not batch_norm
    
    for i in range(stack_num):
        X = DepthwiseConv2D(kernel_size, dilation_rate=dilation_rate, padding='same', 
                            use_bias=bias_flag, name='{}_{}_depthwise'.format(name, i))(X)
        
        if batch_norm:
            X = BatchNormalization(name='{}_{}_depthwise_BN'.format(name, i))(X)

        X = activation_func(name='{}_{}_depthwise_activation'.format(name, i))(X)

        X = Conv2D(channel, (1, 1), padding='same', use_bias=bias_flag, name='{}_{}_pointwise'.format(name, i))(X)
        
        if batch_norm:
            X = BatchNormalization(name='{}_{}_pointwise_BN'.format(name, i))(X)

        X = activation_func(name='{}_{}_pointwise_activation'.format(name, i))(X)
    
    return X

def ASPP_conv(X, channel, activation='ReLU', batch_norm=True, name='aspp'):
    '''
    Atrous Spatial Pyramid Pooling (ASPP).
    
    ASPP_conv(X, channel, activation='ReLU', batch_norm=True, name='aspp')
    
    ----------
    Wang, Y., Liang, B., Ding, M. and Li, J., 2019. Dense semantic labeling 
    with atrous spatial pyramid pooling and decoder for high-resolution remote 
    sensing imagery. Remote Sensing, 11(1), p.20.
    
    Input
    ----------
        X: input tensor.
        channel: number of convolution filters.
        activation: one of the `tensorflow.keras.layers` interface, e.g., ReLU.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
        
    * dilation rates are fixed to `[6, 9, 12]`.
    '''
    
    activation_func = eval(activation)
    bias_flag = not batch_norm

    shape_before = X.get_shape().as_list()
    b4 = GlobalAveragePooling2D(name='{}_avepool_b4'.format(name))(X)
    
    b4 = expand_dims(expand_dims(b4, 1), 1, name='{}_expdim_b4'.format(name))
    
    b4 = Conv2D(channel, 1, padding='same', use_bias=bias_flag, name='{}_conv_b4'.format(name))(b4)
    
    if batch_norm:
        b4 = BatchNormalization(name='{}_conv_b4_BN'.format(name))(b4)
        
    b4 = activation_func(name='{}_conv_b4_activation'.format(name))(b4)
    
    # <----- tensorflow v1 resize.
    b4 = Lambda(lambda X: image.resize(X, shape_before[1:3], method='bilinear', align_corners=True), 
                name='{}_resize_b4'.format(name))(b4)
    
    b0 = Conv2D(channel, (1, 1), padding='same', use_bias=bias_flag, name='{}_conv_b0'.format(name))(X)

    if batch_norm:
        b0 = BatchNormalization(name='{}_conv_b0_BN'.format(name))(b0)
        
    b0 = activation_func(name='{}_conv_b0_activation'.format(name))(b0)
    
    # dilation rates are fixed to `[6, 9, 12]`.
    b_r6 = Sep_CONV_stack(X, channel, kernel_size=3, stack_num=1, activation='ReLU', 
                        dilation_rate=6, batch_norm=True, name='{}_sepconv_r6'.format(name))
    b_r9 = Sep_CONV_stack(X, channel, kernel_size=3, stack_num=1, activation='ReLU', 
                        dilation_rate=9, batch_norm=True, name='{}_sepconv_r9'.format(name))
    b_r12 = Sep_CONV_stack(X, channel, kernel_size=3, stack_num=1, activation='ReLU', 
                        dilation_rate=12, batch_norm=True, name='{}_sepconv_r12'.format(name))
    
    return concatenate([b4, b0, b_r6, b_r9, b_r12])

def CONV_output(X, n_labels, kernel_size=1, activation='Softmax', name='conv_output'):
    '''
    Convolutional layer with output activation.
    
    CONV_output(X, n_labels, kernel_size=1, activation='Softmax', name='conv_output')
    
    Input
    ----------
        X: input tensor.
        n_labels: number of classification label(s).
        kernel_size: size of 2-d convolution kernels. Default is 1-by-1.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interface or 'Sigmoid'.
                    Default option is 'Softmax'.
                    if None is received, then linear activation is applied.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
        
    '''
    
    X = Conv2D(n_labels, kernel_size, padding='same', use_bias=True, name=name)(X)
    
    if activation:
        
        if activation == 'Sigmoid':
            X = Activation('sigmoid', name='{}_activation'.format(name))(X)
            
        else:
            activation_func = eval(activation)
            X = activation_func(name='{}_activation'.format(name))(X)
            
    return X




# from keras_unet_collection.models import unet_3plus_2d
from keras_unet_collection.models import unet_3plus_2d
from tqdm import tqdm
import os.path as osp
import random
import matplotlib.pyplot as plt

input_size = (512, 512, 3)
filter_num_down = [32, 64, 128, 256, 512]
# filter_num_skip = [32, 32, 32, 32]
# filter_num_aggregate = 160
filter_num_skip='auto'
filter_num_aggregate='auto'

stack_num_down=2
stack_num_up=1
activation='ReLU'
output_activation='Sigmoid',
batch_norm=False
pool=True
unpool=True
deep_supervision=True

backbone=None
weights='imagenet'
freeze_backbone=True
freeze_batch_norm=True
name='unet3plus'

use_tf_api = True
                  

from keras_unet_collection import losses

def hybrid_loss(y_true, y_pred):
    loss_focal = losses.focal_tversky(y_true, y_pred, alpha=0.5, gamma=4/3)
    # loss_focal = losses.dice(y_true, y_pred, alpha=0.5, )
    loss_iou = losses.iou_seg(y_true, y_pred)
    
    # (x) 
    #loss_ssim = losses.ms_ssim(y_true, y_pred, max_val=1.0, filter_size=4)
    
    return loss_focal + loss_iou

from glob import glob
import numpy as np
import os
os.environ["CUDA_VISIBLE_DEVICES"] = "0"  

import tensorflow as tf 
print("Num GPUs Available: ", len(tf.config.list_physical_devices('GPU')))
gpus = tf.config.list_physical_devices('GPU')
if gpus:
    try:
        # 메모리 증가 허용
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)
        logical_gpus = tf.config.list_logical_devices('GPU')
        print(f"{len(gpus)} Physical GPUs, {len(logical_gpus)} Logical GPUs")
    except RuntimeError as e:
        print(e)
from tensorflow import keras
import keras_unet_collection.utils as utils

def input_data_process(input_array):
    '''converting pixel vales to [0, 1]'''
    return input_array/255.

def target_data_process(target, num_classes=None, oxford=False):
    if not oxford:
        return keras.utils.to_categorical(target, num_classes=num_classes)
    else:
        return keras.utils.to_categorical(target-1, num_classes=num_classes)

oxford = False
if not oxford:
    input_dir = '/HDD/_projects/benchmark/semantic_segmentation/new_model/datasets/patches_scratch'
    mask_input_dir = '/HDD/_projects/benchmark/semantic_segmentation/new_model/datasets/masks_scratch'
    output_dir = '/HDD/_projects/benchmark/semantic_segmentation/new_model/outputs/unet3p_scratch_'
    n_labels = 2

    import os.path as osp

    if not osp.exists(output_dir):
        os.mkdir(output_dir)

    sample_names = np.array(sorted(glob(input_dir + '/*.bmp')))
    label_names = np.array(sorted(glob(mask_input_dir + '/*.bmp')))
else:
    output_dir = f'/HDD/_projects/benchmark/semantic_segmentation/new_model/outputs/oxford_{use_tf_api}'

    sample_names = np.array(sorted(glob('/HDD/datasets/public/Oxford_IIIT/images/*.jpg')))
    label_names = np.array(sorted(glob('/HDD/datasets/public/Oxford_IIIT/annotations/trimaps/*.png')))
    n_labels = 3

if not osp.exists(output_dir):
    os.mkdir(output_dir)

# model = unet_3plus_2d(input_size, n_labels, filter_num_down, 
#                       filter_num_skip=filter_num_skip, filter_num_aggregate=filter_num_aggregate, 
#                   stack_num_down=stack_num_down, stack_num_up=stack_num_up, activation=activation,
#                   batch_norm=batch_norm, pool=pool, unpool=unpool, deep_supervision=deep_supervision, 
#                   backbone=backbone, weights=weights, freeze_backbone=freeze_backbone, freeze_batch_norm=freeze_batch_norm, 
#                   name='unet3plus')
from unet3p import unet3plus
model = unet3plus


L = len(sample_names)
ind_all = utils.shuffle_ind(L)

if oxford:
    L_train = int(0.2*L)
    L_valid = int(0.02*L)
    L_test = int(0.02*L)
    ind_train = ind_all[:L_train]
    ind_valid = ind_all[L_train:L_train+L_valid]
    ind_test = ind_all[L_train+L_valid:]
else:
    L_train = int(1*L)
    L_valid = int(0.1*L)
    L_test = int(0.1*L)
    ind_train = ind_all[:L_train]
    ind_valid = ind_all[L_train - L_valid:]
    ind_test = ind_all[L_train -L_valid:]
print("Training:validation:testing = {}:{}:{}".format(L_train, L_valid, L_test))

valid_input = input_data_process(utils.image_to_array(sample_names[ind_valid], size=512, channel=3))
valid_target = target_data_process(utils.image_to_array(label_names[ind_valid], size=512, channel=1), n_labels, oxford)

N_epoch = 300 # number of epoches
N_sample = 2 # number of samples per batch

tol = 0 # current early stopping patience
max_tol = 3 # the max-allowed early stopping patience
min_del = 0 # the lowest acceptable loss value reduction 

vis_val = True
cnt = 0

# 옵티마이저 정의
optimizer = tf.keras.optimizers.Adam(learning_rate=1e-5)
loss_weights=[.25, .25, 0.25, 0.25, 1]

# 손실 함수 정의 (여기서는 hybrid_loss를 사용한다고 가정)
def compute_loss(y_true, y_pred):
    # 필요한 손실 계산 수행
    return hybrid_loss(y_true, y_pred)

if use_tf_api:
    model.compile(loss=[hybrid_loss, hybrid_loss, hybrid_loss, hybrid_loss, hybrid_loss],
                  loss_weights=loss_weights,
                  optimizer=optimizer)



# 훈련 루프
total_train_losses, total_val_losses = [], []
for epoch in range(N_epoch):
    
    train_indexes = list(range(L_train))
    random.shuffle(train_indexes)
    check_train_indexes = np.zeros(len(train_indexes))
    train_loss = []
    for step in range(int(L_train/N_sample)):
        print(f"\r train {str(epoch)} ({step}/{int(L_train/N_sample)}) > {str(train_loss[-1]) if len(train_loss) != 0 else ''}", end="")
        
        # 데이터 준비
        train_index = train_indexes[step*N_sample:(step + 1)*N_sample]
        check_train_indexes[train_index] = 1
        train_input = input_data_process(
            utils.image_to_array(sample_names[ind_train][train_index], size=512, channel=3))
        train_target = target_data_process(
                utils.image_to_array(label_names[ind_train][train_index], size=512, channel=1), n_labels, oxford)
        
        if use_tf_api:
            loss = model.train_on_batch([train_input,], 
                                         [train_target, train_target, train_target, train_target, train_target,])
            train_loss.append(loss)
            # preds = model.predict([train_input])
            # pred = np.argmax(preds[0], axis=-1)
            # print('0 ', np.unique(pred))
            # pred = np.argmax(preds[1], axis=-1)
            # print('1 ', np.unique(pred))
            # pred = np.argmax(preds[2], axis=-1)
            # print('2 ', np.unique(pred))
            # pred = np.argmax(preds[3], axis=-1)
            # print('3 ', np.unique(pred))
            # pred = np.argmax(preds[4], axis=-1)
            # print('4 ', np.unique(pred))
        else:
            # with tf.GradientTape() as tape:
            #     y_pred = model([train_input], training=True)
            #     if deep_supervision:
                    
            #         for pred in y_pred:
            #             train_loss.append(compute_loss(train_target, pred).numpy())
            #         weighted_loss = [train_loss[jdx] * loss_weight 
            #                 for jdx, loss_weight in enumerate(loss_weights)]
            #         loss = tf.math.reduce_mean(weighted_loss)
            #     else:
            #         loss = compute_loss(train_target, y_pred)
                    
            with tf.GradientTape() as tape:
                y_pred = model([train_input], training=True)
                if deep_supervision:
                    loss = sum([compute_loss(train_target, y_pred[i]) * loss_weight 
                            for i, loss_weight in enumerate(loss_weights)])
                else:
                    loss = compute_loss(train_target, y_pred)
                    
            grads = tape.gradient(loss, model.trainable_variables)
            optimizer.apply_gradients(zip(grads, model.trainable_variables))
        
    assert sum(check_train_indexes) == len(train_indexes)
    print('train loss: ', np.mean(train_loss))
    total_train_losses.append(np.mean(train_loss))

    if epoch % 3 == 0:
    # if True:
        val_loss = []
        for step in range(int(L_valid/N_sample)):
            print(f"\r val {str(epoch)} ({step}/{int(L_valid/N_sample)}) > {str(val_loss[-1]) if len(val_loss) != 0 else ''}", end="")

            valid_batch = valid_input[N_sample*step:N_sample*(step + 1)]
            target_batch = valid_target[N_sample*step:N_sample*(step + 1)]
            
            if use_tf_api:
                y_pred = model.predict([valid_batch], verbose=False)
                if deep_supervision:
                    val_loss.append(np.mean(hybrid_loss(target_batch, y_pred[-1])))
            else:
                y_pred = model([valid_batch], training=False)
                if deep_supervision:
                    loss = sum([compute_loss(target_batch, y_pred[i]) * loss_weight 
                                for i, loss_weight in enumerate([0.25, 0.25, 0.25, 0.25, 1.0])])
                else:
                    loss = compute_loss(target_batch, y_pred)
                val_loss.append(loss.numpy())

            if vis_val:
                import numpy as np
                import imgviz 
                import cv2
                
                vis_dir = osp.join(output_dir, 'vis')
                if not osp.exists(vis_dir):
                    os.mkdir(vis_dir)
                
                vis_dir = osp.join(vis_dir, str(epoch))
                if not osp.exists(vis_dir):
                    os.mkdir(vis_dir)
                
                for batch_idx in range(len(valid_batch)):
                    vis_img = np.zeros((512, 512*3, 3))
                    vis_img[:, :512, :] = valid_batch[batch_idx]*255
                    color_map = imgviz.label_colormap(50)
                    gt = np.argmax(target_batch[batch_idx], axis=-1).astype(np.uint8)
                    gt = color_map[gt].astype(np.uint8)
                    vis_img[:, 512:512*2, :] = gt 
                    
                    # mask = np.argmax(y_pred[0][batch_idx], axis=-1).astype(np.uint8)
                    # print(np.unique(mask))
                    # mask = np.argmax(y_pred[1][batch_idx], axis=-1).astype(np.uint8)
                    # print(np.unique(mask))
                    # mask = np.argmax(y_pred[2][batch_idx], axis=-1).astype(np.uint8)
                    # print(np.unique(mask))
                    # mask = np.argmax(y_pred[3][batch_idx], axis=-1).astype(np.uint8)
                    # print(np.unique(mask))
                    mask = np.argmax(y_pred[-1][batch_idx], axis=-1).astype(np.uint8)
                    # print(np.unique(mask))
                    mask = color_map[mask].astype(np.uint8)
                    vis_img[:, 512*2:, :] = mask 
                    
                    cv2.imwrite(osp.join(vis_dir, f'{epoch}_{cnt}.bmp'), vis_img)
                    cnt += 1

        val_loss = np.mean(val_loss)
        print('val loss: ', val_loss)
        total_val_losses.append(val_loss)
        model.save(osp.join(output_dir, f'unet3p_{epoch}__.h5'))


    f = open(osp.join(output_dir, 'train_loss.txt'), 'w')
    for _train_loss in total_train_losses:
        f.write(str(_train_loss))
        f.write('\n')
        
    f = open(osp.join(output_dir, 'val_loss.txt'), 'w')
    for _val_loss in total_val_losses:
        f.write(str(_val_loss))
        f.write('\n')
        
    fig = plt.figure()
    plt.plot(total_train_losses)
    plt.savefig(osp.join(output_dir, 'train_loss.jpg'))
    plt.close()
    fig = plt.figure()
    plt.plot(total_val_losses)
    plt.savefig(osp.join(output_dir, 'val_loss.jpg'))
    plt.close()
        
        
